# -*- mode: python; coding: utf-8 -*-
#
# Copyright (C) 2017 <contact@redhat.com>
#
# Author: Loic Dachary <loic@dachary.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import division

import argparse
import collections
import copy
import logging
import textwrap

from crush import Crush
from crush import analyze
from crush import compare
from crush.analyze import Analyze

log = logging.getLogger(__name__)


def top_optimize(args):
    self = args[0]
    return self.optimize_bucket(*args[1:])


class Optimize(object):

    def __init__(self, args, main):
        self.args = args
        self.main = main

    def __getstate__(self):
        return (self.main,)

    def __setstate__(self, state):
        self.main = state[0]
        self.args = self.main.args

    @staticmethod
    def get_parser():
        parser = Analyze.get_parser()
        parser.add_argument(
            '--step',
            help='optimization steps (default infinite)',
            type=int,
        )
        parser.add_argument(
            '--no-forecast',
            dest='with_forecast',
            action='store_false', default=True,
            help='how many steps to completion (default: true)')
        parser.add_argument(
            '--no-positions',
            dest='with_positions',
            action='store_false', default=True,
            help='optimize weiths for each position (default: true)')
        parser.add_argument(
            '--no-multithread',
            dest='multithread',
            action='store_false', default=True,
            help='use multithread when possible (default: true)')
        parser.add_argument(
            '--out-path',
            help='path of the output file')
        return parser

    @staticmethod
    def set_parser(subparsers, arguments):
        parser = Optimize.get_parser()
        arguments(parser)
        subparsers.add_parser(
            'optimize',
            formatter_class=argparse.RawDescriptionHelpFormatter,
            description=textwrap.dedent("""\
            Optimize a crushmap rule

            A rule may not produce the expected distribution if:

            - there are not enough values

            - the weights are not all the same and the replication count is
              two or more

            The optimization of a crush rule consists of modifying the
            weights of each item to get closer to the expected
            distribution.

            A simulation is run with values (--values-count) using the
            rule (--rule) from the crushmap (--crushmap). If the
            observed distribution is different from the expected
            distribution, alternative weights are calculated and
            stored under the --choose-args name. The resulting
            crushmap, including the optimized weights, is saved
            in the --out-path file.

            The optimization runs in parallel, unless the
            --no-multithread flag is given.

            By default weights are optimized for each replication
            level in the range [1,--replication-count]. If the
            --no-positions flag is given, a single set of weights is
            produce and can be used as replacements in crushmaps that
            cannot handle multiple weights per item.

            Optimization is an iterative process that can be stopped
            and resumed later. If the --step flag is not specified,
            the optimization runs until it cannot get closer to the
            desired distribution. If specified, the value of the
            --step flag is the number of items that are moved by
            changing the bucket weights. The optimization will stop if
            more than --step items are moved. For instance --step 1
            stops after the optimization step that moves one value or
            more.

            When --step is specified, the crushmap for the first step
            is written to --out-path and the optimization process
            continues to show how many steps remain before the rule
            cannot be optimized any more. The --no-forecast flag
            forces optimization to stop right after the first step.
            """),
            epilog=textwrap.dedent("""
            Examples:

            crush optimize \\
                  --choose-args optimize --rule replicated_ruleset \\
                  --replication-count 3 --crushmap map.json \\
                  --out-path optimized.json
            2017-05-24 11:12:34,752 default optimizing
            2017-05-24 11:12:39,361 default wants to swap 10 objects
            2017-05-24 11:12:39,362 default will swap 10 objects
            2017-05-24 11:12:39,369 cloud3-1359 optimizing
            2017-05-24 11:12:39,370 cloud3-1360 optimizing
            ...
            """),
            help='Optimize crushmaps',
            parents=[parser],
        ).set_defaults(
            func=Optimize,
        )

    def get_choose_arg(self, crushmap, bucket):
        if 'choose_args' not in crushmap:
            crushmap['choose_args'] = {}
        if self.args.choose_args not in crushmap['choose_args']:
            crushmap['choose_args'][self.args.choose_args] = []
        choose_arg = None
        for element in crushmap['choose_args'][self.args.choose_args]:
            if element['bucket_id'] == bucket['id']:
                choose_arg = element
                break
        if choose_arg is None:
            choose_arg = {'bucket_id': bucket['id']}
            crushmap['choose_args'][self.args.choose_args].append(choose_arg)
        return choose_arg

    def set_choose_arg_position(self, choose_arg, bucket, choose_arg_position):
        if 'weight_set' not in choose_arg:
            choose_arg['weight_set'] = []
        if len(choose_arg['weight_set']) == 0:
            id2weight = collections.OrderedDict(
                [(i['id'], i['weight']) for i in bucket['children']])
            choose_arg['weight_set'] = [list(id2weight.values())]
        if len(choose_arg['weight_set']) <= choose_arg_position:
            last = choose_arg['weight_set'][-1]
            choose_arg['weight_set'].extend(
                [last] * (choose_arg_position - len(choose_arg['weight_set']) + 1))

    def optimize_bucket(self, p, origin_crushmap, bucket):
        if len(bucket.get('children', [])) == 0:
            return None
        if bucket.get('algorithm', 'straw2') != 'straw2':
            raise ValueError(bucket['name'] + ' algorithm is ' + bucket['algorithm'] +
                             ', only straw2 can be optimized')
        log.warning(bucket['name'] + " optimizing")
        crushmap = copy.deepcopy(origin_crushmap)
        if self.args.with_positions:
            for replication_count in range(1, self.args.replication_count + 1):
                log.debug(bucket['name'] + " improving replica " + str(replication_count))
                count = self.optimize_replica(p, origin_crushmap,
                                              crushmap, bucket,
                                              replication_count, replication_count - 1)
        else:
            count = self.optimize_replica(p, origin_crushmap,
                                          crushmap, bucket,
                                          self.args.replication_count, 0)
        if count > 0:
            n = self.main.value_name()
            log.warning(bucket['name'] + " wants to swap " + str(count) + " " + n)
        else:
            log.warning(bucket['name'] + " already optimized")
        return (count, self.get_choose_arg(crushmap, bucket))

    def optimize_replica(self, p, origin_crushmap,
                         crushmap, bucket,
                         replication_count, choose_arg_position):
        a = self.main.clone().constructor(['analyze'] + p)
        a.args.replication_count = replication_count

        parser = compare.Compare.get_parser()
        self.main.hook_compare_args(parser)
        cp = self.main.get_trimmed_argv(parser, self.args)
        compare_instance = self.main.clone().constructor(['compare'] + cp)
        compare_instance.args.replication_count = replication_count
        compare_instance.set_origin_crushmap(origin_crushmap)

        choose_arg = self.get_choose_arg(crushmap, bucket)
        self.set_choose_arg_position(choose_arg, bucket, choose_arg_position)
        id2weight = collections.OrderedDict()
        for pos in range(len(bucket['children'])):
            v = choose_arg['weight_set'][choose_arg_position][pos]
            id2weight[bucket['children'][pos]['id']] = v

        log.info(bucket['name'] + " optimizing replica " + str(replication_count) + " " +
                 str(list(id2weight.values())))
        log.debug(bucket['name'] + " optimizing replica " + str(replication_count) + " " +
                  str(dict(id2weight)))
        c = Crush(backward_compatibility=self.args.backward_compatibility)
        c.parse(crushmap)

        (take, failure_domain) = c.rule_get_take_failure_domain(a.args.rule)
        #
        # initial simulation
        #
        i = a.run_simulation(c, take, failure_domain)
        i = i.reset_index()
        s = i['~name~'] == 'KKKK'  # init to False, there must be a better way
        for item in bucket['children']:
            s |= (i['~name~'] == item['name'])

        previous_delta = None
        improve_tolerance = 10
        no_improvement = 0
        max_iterations = 1000
        from_to_count = 0
        best_weights = list(id2weight.values())
        n = self.main.value_name()
        for iterations in range(max_iterations):
            choose_arg['weight_set'][choose_arg_position] = list(id2weight.values())
            c.parse(crushmap)
            z = a.run_simulation(c, take, failure_domain)
            z = z.reset_index()
            d = z[s].copy()
            d['~delta~'] = d['~' + n + '~'] - d['~expected~']
            d['~delta%~'] = d['~delta~'] / d['~expected~']
            delta = d['~delta~'].abs().sum()
            if previous_delta is not None:
                if previous_delta < delta:
                    no_improvement += 1
                else:
                    previous_delta = delta
                    best_weights = list(id2weight.values())
                    no_improvement = 0
                if no_improvement >= improve_tolerance:
                    log.info("stop because " + str(no_improvement) + " tries")
                    break
            else:
                best_weights = list(id2weight.values())
                previous_delta = delta
            if delta == 0:
                log.info("stop because the distribution is perfect")
                break
            log.info(bucket['name'] + " delta " + str(delta))
            if self.args.step and no_improvement == 0:
                compare_instance.set_destination(c)
                (from_to, in_out) = compare_instance.compare_bucket(bucket)
                from_to_count = sum(map(lambda x: sum(x.values()), from_to.values()))
                in_out_count = sum(map(lambda x: sum(x.values()), in_out.values()))
                log.debug("moved from_to " + str(from_to_count) +
                          " in_out " + str(in_out_count))
                if from_to_count > self.args.step:
                    log.info("stopped because moved " + str(from_to_count) +
                             " --step " + str(self.args.step))
                    break
            d = d.sort_values('~delta~', ascending=False)
            if d.iloc[0]['~delta~'] <= 0 or d.iloc[-1]['~delta~'] >= 0:
                log.info("stop because [" + str(d.iloc[0]['~delta~']) + "," +
                         str(d.iloc[-1]['~delta~']) + "]")
                break
            # there should not be a need to keep the sum of the weights to the same value, they
            # are only used locally for placement and have no impact on the upper weights
            # nor are they derived from the weights from below *HOWEVER* in case of a failure
            # the weights need to be as close as possible from the target weight to limit
            # the negative impact
            shift = int(id2weight[d.iloc[0]['~id~']] * min(0.01, abs(d.iloc[0]['~delta%~'])))
            if shift <= 0:
                log.info("stop because shift is zero")
                break
            log.debug("shift from " + str(d.iloc[0]['~id~']) +
                      " to " + str(d.iloc[-1]['~id~']))
            id2weight[d.iloc[0]['~id~']] -= shift
            id2weight[d.iloc[-1]['~id~']] += shift

        choose_arg['weight_set'][choose_arg_position] = best_weights
        c.parse(crushmap)
        compare_instance.set_destination(c)
        (from_to, in_out) = compare_instance.compare_bucket(bucket)
        from_to_count = sum(map(lambda x: sum(x.values()), from_to.values()))

        if iterations >= max_iterations - 1:
            log.info("stopped after " + str(iterations))
        log.info(bucket['name'] + " replica " + str(replication_count) + " optimized")
        log.info(bucket['name'] + " weights " + str(choose_arg['weight_set'][choose_arg_position]))
        return from_to_count

    def optimize(self, crushmap):
        c = Crush(backward_compatibility=self.args.backward_compatibility)
        c.parse(crushmap)
        crushmap = c.get_crushmap()
        if 'choose_args' not in crushmap:
            crushmap['choose_args'] = {}
            c.parse(crushmap)
        if self.args.choose_args not in crushmap['choose_args']:
            crushmap['choose_args'][self.args.choose_args] = []
            c.parse(crushmap)
        (take, failure_domain) = c.rule_get_take_failure_domain(self.args.rule)

        parser = analyze.Analyze.get_parser()
        self.main.hook_analyze_args(parser)
        p = self.main.get_trimmed_argv(parser, self.args)
        a = self.main.clone().constructor(['analyze'] + p)

        if self.args.multithread:
            from multiprocessing import Pool
            pool = Pool()
        children = [c.find_bucket(take)]
        total_count = 0
        over_step = False
        n = self.main.value_name()
        while not over_step and len(children) > 0:
            a = [(self, p, c.get_crushmap(), item) for item in children]
            if self.args.multithread:
                r = list(pool.map(top_optimize, a))
            else:
                r = list(map(top_optimize, a))
            for i in range(len(children)):
                if r[i] is None:
                    continue
                (count, choose_arg) = r[i]
                total_count += count
                c.update_choose_args(self.args.choose_args, [choose_arg])
                log.info(children[i]['name'] + " weights updated with " + str(choose_arg))
                if self.args.step and count > 0:
                    log.warning(children[i]['name'] + " will swap " +
                                str(count) + " " + n)
                over_step = self.args.step and total_count > self.args.step
                if over_step:
                    break
            nc = []
            for item in children:
                nc.extend(item.get('children', []))
            # fail if all children are not of the same type
            children = nc
        return (total_count, c.get_crushmap())

    def run(self):
        if not self.args.crushmap:
            raise Exception("missing --crushmap")
        crushmap = self.main.convert_to_crushmap(self.args.crushmap)
        if not self.args.choose_args:
            raise Exception("missing --choose-args")
        (count, crushmap) = self.optimize(crushmap)
        self.main.crushmap_to_file(crushmap)
        if self.args.step and self.args.with_forecast:
            log.warning("the optimized crushmap was written to " + self.args.out_path)
            log.warning("now running simulation of the next steps")
            log.warning("this can be disabled with --no-forecast")
            step = 2
            n = self.main.value_name()
            while count > 0:
                (count, crushmap) = self.optimize(crushmap)
                log.warning("step " + str(step) + " moves " + str(count) + " " + n)
                step += 1



# -*- mode: python; coding: utf-8 -*-
#
# Copyright (C) 2017 <contact@redhat.com>
#
# Author: Loic Dachary <loic@dachary.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
import pytest

from crush.ceph import CephCrushmapConverter, CephCrush


class TestCephCrush(object):

    def test_convert_to_crushmap(self):
        c = CephCrush()
        crushmap = {}
        assert crushmap == c._convert_to_crushmap(crushmap)
        crushmap = c._convert_to_crushmap("tests/sample-crushmap.json")
        assert 'trees' in crushmap
        crushmap = c._convert_to_crushmap("tests/sample-ceph-crushmap.txt")
        assert 'trees' in crushmap
        crushmap = c._convert_to_crushmap("tests/sample-ceph-crushmap.crush")
        assert 'trees' in crushmap
        crushmap = c._convert_to_crushmap("tests/sample-ceph-crushmap.json")
        assert 'trees' in crushmap
        with pytest.raises(ValueError) as e:
            crushmap = c._convert_to_crushmap("tests/sample-bugous-crushmap.json")
        assert "Expecting property name" in str(e.value)


class TestCephCrushmapConverter(object):

    def test_recover_choose_args(self):
        ceph = {
            'buckets': [
                {
                    'name': 'SELF-target-weight',
                    'id': -10,
                    'items': [
                        {'weight': 1 * 0x10000, 'id': 1},
                        {'weight': 2 * 0x10000, 'id': 2}
                    ]
                },
                {
                    'name': 'SELF',
                    'id': -1,
                    'items': [
                        {'weight': 10 * 0x10000, 'id': 1},
                        {'weight': 20 * 0x10000, 'id': 2}
                    ]
                },
            ]
        }
        CephCrushmapConverter.recover_choose_args(ceph)
        expected = {
            'choose_args': {' placeholder ': [{'bucket_id': -1, 'weight_set': [[10, 20]]}]},
            'buckets': [
                {
                    'name': 'SELF',
                    'id': -1,
                    'items': [
                        {'weight': 1 * 0x10000, 'id': 1},
                        {'weight': 2 * 0x10000, 'id': 2}
                    ]
                },
            ]
        }
        assert expected == ceph

    def test_recover_choose_args_added(self):
        ceph = {
            'buckets': [
                {
                    'name': 'SELF-target-weight',
                    'id': -10,
                    'items': [
                        {'weight': 1 * 0x10000, 'id': 1},
                    ]
                },
                {
                    'name': 'SELF',
                    'id': -1,
                    'items': [
                        {'weight': 10 * 0x10000, 'id': 1},
                        {'weight': 2 * 0x10000, 'id': 2}
                    ]
                },
            ]
        }
        CephCrushmapConverter.recover_choose_args(ceph)
        expected = {
            'choose_args': {' placeholder ': [{'bucket_id': -1, 'weight_set': [[10, 0]]}]},
            'buckets': [
                {
                    'name': 'SELF',
                    'id': -1,
                    'items': [
                        {'weight': 1 * 0x10000, 'id': 1},
                        {'weight': 2 * 0x10000, 'id': 2}
                    ]
                },
            ]
        }
        assert expected == ceph

    def test_recover_choose_args_removed(self):
        ceph = {
            'buckets': [
                {
                    'name': 'SELF-target-weight',
                    'id': -10,
                    'items': [
                        {'weight': 1 * 0x10000, 'id': 1},
                        {'weight': 2 * 0x10000, 'id': 2}
                    ]
                },
                {
                    'name': 'SELF',
                    'id': -1,
                    'items': [
                        {'weight': 20 * 0x10000, 'id': 2}
                    ]
                },
            ]
        }
        CephCrushmapConverter.recover_choose_args(ceph)
        expected = {
            'choose_args': {' placeholder ': [{'bucket_id': -1, 'weight_set': [[20]]}]},
            'buckets': [
                {
                    'name': 'SELF',
                    'id': -1,
                    'items': [
                        {'weight': 2 * 0x10000, 'id': 2}
                    ]
                },
            ]
        }
        assert expected == ceph

# Local Variables:
# compile-command: "cd .. ; tox -e py27 -- -s -vv tests/test_ceph_crush.py"
# End:



# -*- mode: python; coding: utf-8 -*-
#
# Copyright (C) 2017 <contact@redhat.com>
#
# Author: Loic Dachary <loic@dachary.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
import copy
import pytest # noqa needed for caplog

from crush import Crush


class TestCrush(object):

    def build_crushmap(self):
        crushmap = {
            "trees": [
                {
                    "type": "root",
                    "id": -1,
                    "name": "dc1",
                    "children": [],
                }
            ],
            "rules": {
                "data": [
                    ["take", "dc1"],
                    ["chooseleaf", "firstn", 0, "type", "host"],
                    ["emit"]
                ],
            }
        }
        crushmap['trees'][0]['children'].extend([
            {
                "type": "host",
                "id": -(i + 2),
                "name": "host%d" % i,
                "children": [
                    {"id": (2 * i), "name": "device%02d" % (2 * i), "weight": 1},
                    {"id": (2 * i + 1), "name": "device%02d" % (2 * i + 1), "weight": 2},
                ],
            } for i in range(0, 10)
        ])
        return crushmap

    def test_map(self):
        crushmap = self.build_crushmap()
        c = Crush(verbose=1)
        assert c.parse(crushmap)
        assert len(c.map(rule="data", value=1234, replication_count=1,
                         weights={}, choose_args=[])) == 1

    def test_get_item_by_(self):
        crushmap = self.build_crushmap()
        c = Crush(verbose=1)
        assert c.parse(crushmap)
        assert c.get_item_by_id(-2)['name'] == 'host0'
        assert c.get_item_by_name('host0')['id'] == -2

    def test_convert_to_crushmap(self):
        c = Crush()
        crushmap = {}
        assert crushmap == c._convert_to_crushmap(crushmap)
        crushmap = c._convert_to_crushmap("tests/sample-crushmap.json")
        assert 'trees' in crushmap
        with pytest.raises(ValueError) as e:
            crushmap = c._convert_to_crushmap("tests/sample-bugous-crushmap.json")
        assert "Expecting property name" in str(e.value)

    def test_parse_weights_file(self):

        # Test Simple weights file
        weights = Crush.parse_weights_file(open("tests/weights.json"))
        assert weights == {"osd.0": 0.0, "osd.2": 0.5}

        # Test OSDMap
        weights = Crush.parse_weights_file(open("tests/ceph/osdmap.json"))
        assert weights == {"osd.1": 0.95}

        with pytest.raises(AssertionError):
            Crush.parse_weights_file(open("tests/ceph/weights-notfloat.json"))
        with pytest.raises(Exception):
            Crush.parse_weights_file(open("tests/ceph/osdmap-invalid.json"))
        with pytest.raises(AssertionError):
            Crush.parse_weights_file(open("tests/sample-ceph-crushmap.txt"))

    def test_merge_split_choose_args(self):
        c = Crush()
        split = {
            'choose_args': {
                'a': [
                    {'bucket_id': -3},
                    {'bucket_id': -2},
                    {'bucket_id': -1},
                ],
                'b': [
                    {'bucket_id': -1},
                ]
            },
            'trees': [
                {
                    'id': -1,
                    'children': [{
                        'id': -3,
                        'children': [{'id': -4}]
                    }],
                },
                {'id': -2},
            ]
        }
        merged = {
            'trees': [
                {
                    'id': -1,
                    'children': [{
                        'id': -3,
                        'children': [{'id': -4}],
                        'choose_args': {'a': {'bucket_id': -3}},
                    }],
                    'choose_args': {'a': {'bucket_id': -1}, 'b': {'bucket_id': -1}},
                },
                {'id': -2, 'choose_args': {'a': {'bucket_id': -2}}}
            ]
        }

        c.crushmap = copy.deepcopy(split)
        assert c._merge_choose_args()
        assert merged == c.crushmap
        # do nothing if no choose_args
        assert c._merge_choose_args() is False
        c._split_choose_args()
        assert split == c.crushmap

    def test_filter_real(self):
        name = 'cloud6-1429'
        c = Crush()
        c.parse('tests/test_crush_filter.json')
        crushmap = c.get_crushmap()
        optimize = sorted(crushmap['choose_args']['optimize'], key=lambda v: v['bucket_id'])
        assert 3 == len(optimize)
        assert -1 == optimize[2]['bucket_id']
        assert 7 == len(optimize[2]['weight_set'][0])
        bucket = c.find_bucket(name)
        assert name == bucket['name']
        c.filter(lambda x: x.get('name') != name, crushmap['trees'][0])
        optimize = crushmap['choose_args']['optimize']
        assert 2 == len(optimize)
        assert -1 == optimize[1]['bucket_id']
        assert 6 == len(optimize[1]['weight_set'][0])
        assert c.find_bucket(name) is None

    def test_filter_basic(self):
        root = {
            'name': 'root',
            'id': -5,
            'children': [
                {'name': 'bucket2', 'id': -3},
                {'name': 'bucket1', 'id': -2, 'children': [{'id': 1}, {'id': 2}, {'id': 4}]},
                {'name': 'bucket3', 'id': -1, 'children': [{'id': 5}, {'id': 6}, {'id': 7}]},
            ]
        }
        expected_root = {
            'name': 'root',
            'id': -5,
            'children': [
                {'name': 'bucket2', 'id': -3},
                {'name': 'bucket1', 'id': -2, 'children': [{'id': 1}]},
                {'name': 'bucket3', 'id': -1, 'children': [{'id': 5}, {'id': 7}]},
            ]
        }
        choose_args = [
            {'bucket_id': -2, 'ids': [11, 12, 14],
             'weight_set': [[11.0, 12.0, 14.0]]},
            {'bucket_id': -1, 'ids': [15, 16, 17]},
        ]
        expected_choose_args = [
            {'bucket_id': -2, 'ids': [11],
             'weight_set': [[11.0]]},
            {'bucket_id': -1, 'ids': [15, 17]},
        ]
        c = Crush()
        c.crushmap = {}
        c.crushmap['trees'] = [root]
        c.crushmap['choose_args'] = {
            "one": choose_args,
            "empty": [],
        }

        def fun(x):
            if x.get('id') in (2, 4, 6):
                return False
            return True

        c.filter(fun, root)
        assert expected_root == root
        assert expected_choose_args == choose_args
        assert "empty" in c.crushmap['choose_args']

    def test_collect_buckets_by_type(self):
        children = [
            {
                'type': 'host',
                'name': 'host0',
            },
            {
                'type': 'rack',
                'children': [
                    {
                        'name': 'host1',
                        'type': 'host',
                    },
                    {
                        'name': 'host2',
                        'type': 'other',
                    }
                ],
            }
        ]
        expected = [{'name': 'host0', 'type': 'host'}, {'name': 'host1', 'type': 'host'}]
        assert expected == Crush.collect_buckets_by_type(children, 'host')

    def test_update_choose_args(self):
        c = Crush()
        c.crushmap = {}
        choose_args = [{"bucket_id": -1}]
        c.update_choose_args('name', choose_args)
        assert choose_args == c.crushmap['choose_args']['name']

        c.update_choose_args('other_name', choose_args)
        assert choose_args == c.crushmap['choose_args']['other_name']

        choose_args = [{"bucket_id": -2}]
        c.update_choose_args('name', choose_args)
        expected = [{'bucket_id': -2}, {'bucket_id': -1}]
        assert expected == c.crushmap['choose_args']['name']

        choose_args[0]['modified'] = True
        c.update_choose_args('name', choose_args)
        expected = [{'bucket_id': -2, 'modified': True}, {'bucket_id': -1}]
        assert expected == c.crushmap['choose_args']['name']


# Local Variables:
# compile-command: "cd .. ; tox -e py27 -- -s -vv tests/test_crush.py"
# End:



# -*- mode: python; coding: utf-8 -*-
#
# Copyright (C) 2017 <contact@redhat.com>
#
# Author: Loic Dachary <loic@dachary.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import division

import argparse
import collections
import copy
import json
import logging
import re
import struct
import textwrap

from crush import main
from crush.ceph import convert
from crush import Crush, LibCrush

log = logging.getLogger(__name__)


class MappingError(Exception):
    pass


class UnsupportedError(Exception):
    pass


class HealthError(Exception):
    pass


class CephReport(object):

    def parse_report(self, report):
        if report['health']['overall_status'] != 'HEALTH_OK':
            raise HealthError("expected health overall_status == HEALTH_OK but got " +
                              report['health']['overall_status'] + "instead")
        crushmap = CephCrushmapConverter().parse_ceph(report['crushmap'],
                                                      recover_choose_args=False)
        mappings = collections.defaultdict(lambda: {})
        for pg_stat in report['pgmap']['pg_stats']:
            mappings[pg_stat['pgid']] = pg_stat['acting']

        ruleset2name = {}
        for rule in crushmap['private']['rules']:
            ruleset2name[rule['ruleset']] = rule['rule_name']

        c = LibCrush(backward_compatibility=True)
        c.parse(crushmap)

        name2id = {}

        def collect_items(children):
            for child in children:
                if 'id' in child:
                    name2id[child['name']] = child['id']
                collect_items(child.get('children', []))
        collect_items(crushmap['trees'])

        weights = Crush.parse_osdmap_weights(report['osdmap'])

        for osd in report['osdmap']["osds"]:
            if osd["primary_affinity"] != 1.0:
                raise UnsupportedError(
                    "osd." + str(osd["osd"]) + " primary affinity is != 1.0")

        failed_mapping = False
        for pool in report['osdmap']['pools']:
            if pool['type'] != 1:
                raise UnsupportedError(
                    "pool " + pool['pool_name'] + " is type " + str(pool['type']) +
                    " is not supported, only type == 1 (replicated)")
            if pool['object_hash'] != 2:
                raise UnsupportedError(
                    "pool " + pool['pool_name'] + " object_hash " + str(pool['object_hash']) +
                    " is not supported, only object_hash == 2 (rjenkins)")
            if pool['flags_names'] != 'hashpspool':
                raise UnsupportedError(
                    "pool " + pool['pool_name'] + " has flags_names " +
                    "'" + str(pool['flags_names']) + "'" +
                    " is no supported, only hashpspool")
            ruleset = pool['crush_ruleset']
            if str(ruleset) in crushmap.get('choose_args', {}):
                choose_args = str(ruleset)
            else:
                choose_args = None
            rule = ruleset2name[ruleset]
            size = pool['size']
            log.info("verifying pool {} pg_num {} pgp_num {}".format(
                pool['pool'], pool['pg_num'], pool['pg_placement_num']))
            values = LibCrush().ceph_pool_pps(pool['pool'],
                                              pool['pg_num'],
                                              pool['pg_placement_num'])
            kwargs = {
                "rule": str(rule),
                "replication_count": size,
            }
            if choose_args:
                kwargs["choose_args"] = choose_args
            if weights:
                kwargs["weights"] = weights
            for (name, pps) in values.items():
                if name not in mappings:
                    failed_mapping = True
                    log.error(name + " is not in pgmap")
                    continue
                kwargs["value"] = pps
                mapped = c.map(**kwargs)
                osds = [name2id[x] for x in mapped]
                if osds != mappings[name]:
                    failed_mapping = True
                    log.error("{} map to {} instead of {}".format(
                        name, osds, mappings[name]))
                    continue
        if failed_mapping:
            raise MappingError("some mapping failed, please file a bug at "
                               "http://libcrush.org/main/python-crush/issues/new")

        crushmap = CephCrushmapConverter().parse_ceph(report['crushmap'],
                                                      recover_choose_args=True)
        crushmap['private']['pools'] = report['osdmap']['pools']

        (version, rest) = report['version'].split('.', 1)
        crushmap['private']['version'] = int(version)

        return crushmap


class CephCrushmapConverter(object):

    def convert_item(self, item, ceph):
        if item['id'] >= 0:
            return {
                "weight": item['weight'],
                "id": item['id'],
                "name": ceph['id2name'][item['id']],
            }
        else:
            return {
                "weight": item['weight'],
                "reference_id": item['id'],
            }

    def convert_bucket(self, bucket, ceph):
        b = {
            "weight": bucket['weight'],
            "id": bucket['id'],
            "name": bucket['name'],
            "algorithm": bucket['alg'],
            "type": bucket['type_name'],
        }
        items = bucket.get('items', [])
        if items:
            children = []
            last_pos = -1
            for item in items:
                # the actual pos value does not change the mapping
                # when there is an empty item (we do not store pos)
                # but the order of the items is important and we
                # need to assert that the list is ordered
                assert last_pos < item['pos']
                last_pos = item['pos']
                children.append(self.convert_item(item, ceph))
            b['children'] = children
        return b

    def collect_items(self, children, ceph):
        for child in children:
            if 'id' in child:
                ceph['id2item'][child['id']] = child
            self.collect_items(child.get('children', []), ceph)

    def dereference(self, children, ceph):
        for i in range(len(children)):
            child = children[i]
            if 'reference_id' in child:
                id = child['reference_id']
                new_child = copy.copy(ceph['id2item'][id])
                new_child['weight'] = child['weight']
                ceph['is_child'].add(id)
                children[i] = new_child
            self.dereference(child.get('children', []), ceph)

    @staticmethod
    def recover_choose_args(ceph):
        name2id = {}
        id2bucket = {}
        for bucket in ceph['buckets']:
            name2id[bucket['name']] = bucket['id']
            id2bucket[bucket['id']] = bucket
        buckets = []
        name2target_weights = {}
        has_target_weight = False
        for bucket in ceph['buckets']:
            log.debug(str(bucket))
            if bucket['name'].endswith('-target-weight'):
                has_target_weight = True
                name = bucket['name'][:-14]
                target_weights = {}
                for i in bucket['items']:
                    if i['id'] < 0:
                        c = id2bucket[i['id']]
                        assert c['name'].endswith('-target-weight')
                        child_name = c['name'][:-14]
                        id = name2id[child_name]
                    else:
                        id = i['id']
                    target_weights[id] = i['weight']
                name2target_weights[name] = target_weights
            else:
                buckets.append(bucket)
        if not has_target_weight:
            return
        choose_args = []
        for bucket in buckets:
            if bucket['name'] in name2target_weights:
                target_weights = name2target_weights[bucket['name']]
                weight_set = []
                for child in bucket['items']:
                    if child['id'] in target_weights:
                        weight_set.append(child['weight'] / 0x10000)
                        child['weight'] = target_weights[child['id']]
                    else:
                        weight_set.append(0)
                choose_args.append({
                    'bucket_id': bucket['id'],
                    'weight_set': [weight_set],
                })
        ceph['buckets'] = buckets
        ceph['choose_args'] = {" placeholder ": choose_args}

    def convert_buckets(self, ceph, recover_choose_args):
        ceph['is_child'] = set()
        ceph['id2item'] = {}
        if recover_choose_args:
            self.recover_choose_args(ceph)
        converted = [self.convert_bucket(r, ceph) for r in ceph['buckets']]
        self.collect_items(converted, ceph)
        self.dereference(converted, ceph)
        return list(filter(lambda c: c['id'] not in ceph['is_child'], converted))

    def convert_rule(self, ceph_rule, ceph):
        name = ceph_rule['rule_name']
        rule = []
        for ceph_step in ceph_rule['steps']:
            if 'opcode' in ceph_step:
                if ceph_step['opcode'] in (10, 11, 12, 13):
                    id2name = {
                        10: 'set_choose_local_tries',
                        11: 'set_choose_local_fallback_tries',
                        12: 'set_chooseleaf_vary_r',
                        13: 'set_chooseleaf_stable',
                    }
                    step = [id2name[ceph_step['opcode']], ceph_step['arg1']]
                else:
                    assert 0, "unexpected rule opcode " + str(ceph_step['opcode'])
            elif 'op' in ceph_step:
                if ceph_step['op'] == 'take':
                    step = [ceph_step['op'], ceph_step['item_name']]
                elif ceph_step['op'] in ('chooseleaf_firstn',
                                         'choose_firstn',
                                         'chooseleaf_indep',
                                         'choose_indep'):
                    (choose, how) = ceph_step['op'].split('_')
                    if ceph['type2id'][ceph_step['type']] == 0:
                        type = 0
                    else:
                        type = ceph_step['type']
                    step = [choose, how, ceph_step['num'], 'type', type]
                elif ceph_step['op'] in ('set_choose_local_tries',
                                         'set_choose_local_fallback_tries',
                                         'set_chooseleaf_vary_r',
                                         'set_chooseleaf_stable',
                                         'set_choose_tries',
                                         'set_chooseleaf_tries'):
                    step = [ceph_step['op'], ceph_step['num']]
                elif ceph_step['op'] == 'emit':
                    step = ['emit']
                elif ceph_step['op'] == 'noop':
                    pass
                else:
                    assert 0, "unexpected rule op " + str(ceph_step['op'])
            else:
                assert 0, "no op or opcode found"
            rule.append(step)
        return (name, rule)

    def convert_tunables(self, tunables):
        known = set([
            'choose_local_tries',
            'choose_local_fallback_tries',
            'chooseleaf_vary_r',
            'chooseleaf_stable',
            'chooseleaf_descend_once',
            'straw_calc_version',

            'choose_total_tries',
        ])
        out = {}
        for (k, v) in tunables.items():
            if k in known:
                out[k] = v
        return out

    def convert_choose_args(self, ceph):
        choose_args_map = copy.deepcopy(ceph['choose_args'])
        for (name, choose_args) in choose_args_map.items():
            for choose_arg in choose_args:
                if 'weight_set' in choose_arg:
                    choose_arg['weight_set'] = [
                        [int(x * 0x10000) for x in weights]
                        for weights in choose_arg['weight_set']
                    ]
        return choose_args_map

    def parse_ceph(self, ceph, recover_choose_args):
        ceph['id2name'] = {d['id']: d['name'] for d in ceph['devices']}
        ceph['type2id'] = {t['name']: t['type_id'] for t in ceph['types']}

        j = {
            'private': {},
        }

        j['types'] = ceph['types']

        j['trees'] = self.convert_buckets(ceph, recover_choose_args)

        j['rules'] = {}
        for ceph_rule in ceph['rules']:
            (name, rule) = self.convert_rule(ceph_rule, ceph)
            j['rules'][name] = rule
        j['private']['rules'] = ceph['rules']

        j['tunables'] = self.convert_tunables(ceph['tunables'])

        j['private']['tunables'] = ceph['tunables']

        if 'choose_args' in ceph:
            j['choose_args'] = self.convert_choose_args(ceph)

        return j


class CephCrush(Crush):

    #
    # reading a crushmap from a file
    #

    @staticmethod
    def _is_ceph_file(something):
        fmt = "I"
        crush_magic = 0x00010000
        head = open(something, mode='rb').read(struct.calcsize(fmt))
        if struct.unpack(fmt, head)[0] == crush_magic:
            return True
        content = open(something).read()
        if (re.search("^device ", content, re.MULTILINE) and
                re.search("^type ", content, re.MULTILINE)):
            return True
        return False

    @staticmethod
    def _convert_from_file(something):
        if CephCrush._is_ceph_file(something):
            crushmap = LibCrush().ceph_read(something)
            return (json.loads(crushmap), 'ceph-json')
        else:
            with open(something) as f_json:
                crushmap = json.load(f_json)
                log.debug("_detect_file_format: valid json file")
                if 'devices' in crushmap:  # Ceph json format
                    return (crushmap, 'ceph-json')
                elif 'cluster_fingerprint' in crushmap:
                    return (crushmap, 'ceph-report')
                return (crushmap, 'python-crush-json')

    @staticmethod
    def _convert_to_dict(something):
        if type(something) in (dict, collections.OrderedDict):
            if 'devices' in something:
                return (something, 'ceph-json')
            elif 'cluster_fingerprint' in something:
                return (something, 'ceph-report')
            return (something, 'python-crush-json')
        else:
            return CephCrush._convert_from_file(something)

    def _convert_to_crushmap(self, something):
        (something, format) = CephCrush._convert_to_dict(something)
        if format == 'ceph-json':
            crushmap = CephCrushmapConverter().parse_ceph(something,
                                                          recover_choose_args=True)
        elif format == 'ceph-report':
            crushmap = CephReport().parse_report(something)
        else:
            crushmap = something
        return crushmap

    #
    # writing a crushmap to a file
    #

    @staticmethod
    def choose_args_int_index(crushmap):
        if 'choose_args' not in crushmap:
            return crushmap
        crushmap['choose_args'] = {
            int(k): v for (k, v) in crushmap['choose_args'].items()
        }
        return crushmap

    def ceph_version_compat(self):
        #
        # sanity checks
        #
        if self.c.ceph_incompat():
            raise Exception("choose_args cannot be encoded for a version lower than luminous")

        self._merge_choose_args()
        #
        # create the shadow trees with the target weights
        #
        self.max_bucket_id = min(self._id2item.keys())

        def rename(bucket):
            if 'children' not in bucket:
                return
            self.max_bucket_id -= 1
            bucket['id'] = self.max_bucket_id
            bucket['name'] += '-target-weight'
            if 'choose_args' in bucket:
                del bucket['choose_args']
            for child in bucket.get('children', []):
                rename(child)
        shadow_trees = copy.deepcopy(self.crushmap['trees'])
        for tree in shadow_trees:
            rename(tree)

        #
        # override the target weights with the weight set
        #
        def reweight(bucket):
            if 'children' not in bucket:
                return
            children = bucket['children']
            if 'choose_args' in bucket:
                choose_arg = next(iter(bucket['choose_args'].values()))
                weight_set = choose_arg['weight_set'][0]
                for i in range(len(children)):
                    children[i]['weight'] = weight_set[i]
                del bucket['choose_args']
            for child in children:
                reweight(child)
        for tree in self.crushmap['trees']:
            reweight(tree)

        self.crushmap['trees'].extend(shadow_trees)

    def transform_to_write(self, version):
        if 'choose_args' not in self.crushmap:
            return False
        self.choose_args_int_index(self.crushmap)
        self.parse(self.crushmap)
        if version >= 'luminous':
            return True
        self.ceph_version_compat()
        self.parse(self.crushmap)
        return True

    def to_file(self, path, format, version):
        if format == 'python-json':
            super(CephCrush, self).to_file(path)
        else:
            self.transform_to_write(version)
            self.c.ceph_write(path, format, self.crushmap.get('private'))


class Ceph(main.Main):

    def __init__(self):
        super(Ceph, self).__init__()

        self.parser.add_argument(
            '--no-backward-compatibility',
            dest='backward_compatibility',
            action='store_false', default=True,
            help='do not allow backward compatibility tunables (default: allowed)')

        convert.Convert.set_parser(self.subparsers, self.hook_convert_args)

    def create_parser(self):
        self.parser = argparse.ArgumentParser(
            formatter_class=argparse.RawDescriptionHelpFormatter,
            description=textwrap.dedent("""\
            Ceph crush compare and analyze
            """),
            epilog=textwrap.dedent("""
            """),
        )

    def clone(self):
        return Ceph()

    def hook_common_args(self, parser):
        parser.add_argument(
            '--pool',
            help='pool',
            type=int)

        parser.add_argument(
            '--pg-num',
            help='pg-num',
            type=int)

        parser.add_argument(
            '--pgp-num',
            help='pgp-num',
            type=int)

    formats = ('txt', 'json', 'python-json', 'crush')

    def in_args(self, parser):
        parser.add_argument(
            '--in-path',
            help='path of the input file')
        parser.add_argument(
            '--in-format',
            choices=Ceph.formats,
            help='format of the input file')

    def out_args(self, parser):
        parser.add_argument(
            '--out-format',
            choices=Ceph.formats,
            default='crush',
            help='format of the output file')
        versions = ('h', 'hammer',
                    'j', 'jewel',
                    'k', 'kraken',
                    'l', 'luminous',
                    'm',
                    'n')
        parser.add_argument(
            '--out-version',
            choices=versions,
            default='luminous',
            help='version of the output file (default luminous)')

    def hook_convert_args(self, parser):
        self.in_args(parser)
        self.out_args(parser)

    def hook_analyze_args(self, parser):
        self.hook_common_args(parser)

    def hook_compare_args(self, parser):
        self.hook_common_args(parser)

    def hook_optimize_args(self, parser):
        self.hook_common_args(parser)
        self.out_args(parser)

    def hook_create_values(self):
        if self.args.pool is not None:
            return LibCrush().ceph_pool_pps(self.args.pool, self.args.pg_num, self.args.pgp_num)
        else:
            return super(Ceph, self).hook_create_values()

    def value_name(self):
        return 'PGs'

    def get_ceph_version(self, crushmap):
        if 'version' in crushmap['private']:
            return chr(crushmap['private']['version'] - 1 + ord('a'))
        return 'l'

    def has_compat_crushmap(self, crushmap):
        return crushmap.get('choose_args', {}).get(" placeholder ") is not None

    def get_compat_choose_args(self, crushmap):
        #
        # if converting from a pre-Luminous encoded crushmap
        #
        if not self.has_compat_crushmap(crushmap):
            return None
        else:
            assert 'private' in crushmap
            assert 'pools' in crushmap['private']
            assert 1 == len(crushmap['private']['pools'])
            pool = crushmap['private']['pools'][0]
            return str(pool['pool'])

    def set_analyze_args(self, crushmap):
        if 'private' not in crushmap:
            return None
        if 'pools' not in crushmap['private']:
            return None

        compat_pool = self.get_compat_choose_args(crushmap)
        if (compat_pool is not None and
                (self.args.pool is None or self.args.pool == int(compat_pool))):
            self.args.pool = int(compat_pool)
            self.argv.append('--pool=' + str(self.args.pool))
            self.args.choose_args = str(self.args.pool)
            self.argv.append('--choose-args=' + self.args.choose_args)

        for pool in crushmap['private']['pools']:
            if pool['pool'] == self.args.pool:
                self.args.replication_count = pool['size']
                self.argv.append('--replication-count=' + str(pool['size']))
                self.args.pg_num = pool['pg_num']
                self.argv.append('--pg-num=' + str(pool['pg_num']))
                self.args.pgp_num = pool['pg_placement_num']
                self.argv.append('--pgp-num=' + str(pool['pg_placement_num']))
                for rule in crushmap['private']['rules']:
                    if rule['ruleset'] == pool['crush_ruleset']:
                        self.args.rule = str(rule['rule_name'])
                        self.argv.append('--rule=' + str(rule['rule_name']))

        if crushmap.get('choose_args', {}).get(str(self.args.pool)):
            self.args.choose_args = str(self.args.pool)
            self.argv.append('--choose-args=' + self.args.choose_args)

        log.info('argv = ' + " ".join(self.argv))

        return self.args.choose_args

    def set_optimize_args(self, crushmap):
        if 'version' not in crushmap['private']:
            return self.args.choose_args
        self.args.out_version = self.get_ceph_version(crushmap)
        self.argv.append('--out-version=' + self.args.out_version)

        if self.args.out_version < 'luminous':
            self.args.with_positions = False
            self.argv.append('--no-positions')

        if not hasattr(self.args, 'pool'):
            return self.args.choose_args
        if 'private' not in crushmap:
            return self.args.choose_args
        if 'pools' not in crushmap['private']:
            return self.args.choose_args

        if self.args.choose_args is None:
            self.args.choose_args = str(self.args.pool)
            self.argv.append('--choose-args=' + self.args.choose_args)

        log.warning('argv = ' + " ".join(self.argv))

        return self.args.choose_args

    def set_compat_choose_args(self, c, crushmap, choose_args_name):

        if not self.has_compat_crushmap(crushmap):
            return

        assert choose_args_name

        choose_args = crushmap['choose_args']
        choose_args[choose_args_name] = choose_args[' placeholder ']
        del choose_args[' placeholder ']
        c.parse(crushmap)

    def convert_to_crushmap(self, crushmap):
        c = CephCrush(verbose=self.args.debug,
                      backward_compatibility=self.args.backward_compatibility)
        c.parse(crushmap)
        crushmap = c.get_crushmap()
        if self.args.func.__name__ == 'Analyze':
            choose_args_name = self.set_analyze_args(crushmap)
        elif self.args.func.__name__ == 'Optimize':
            self.set_analyze_args(crushmap)
            choose_args_name = self.set_optimize_args(crushmap)
        elif self.args.func.__name__ == 'Convert':
            choose_args_name = self.get_compat_choose_args(crushmap)
        elif self.args.func.__name__ == 'Compare':
            choose_args_name = None
        else:
            raise Exception('Unexpected func=' + str(self.args.func.__name__))
        self.set_compat_choose_args(c, crushmap, choose_args_name)

        return crushmap

    def crushmap_to_file(self, crushmap):
        c = CephCrush(verbose=self.args.debug,
                      backward_compatibility=self.args.backward_compatibility)
        c.parse(crushmap)
        c.to_file(self.args.out_path, self.args.out_format, self.args.out_version)



# -*- mode: python; coding: utf-8 -*-
#
# Copyright (C) 2017 <contact@redhat.com>
#
# Author: Loic Dachary <loic@dachary.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import division

import argparse
import collections
import copy
import logging
import textwrap
import pandas as pd
import numpy as np

from crush import Crush

log = logging.getLogger(__name__)


class BadMapping(Exception):
    pass


class Analyze(object):

    def __init__(self, args, main):
        self.args = args
        self.main = main

    @staticmethod
    def get_parser():
        parser = argparse.ArgumentParser(
            add_help=False,
            conflict_handler='resolve',
        )
        replication_count = 3
        parser.add_argument(
            '--replication-count',
            help=('number of devices to map (default: %d)' % replication_count),
            type=int,
            default=replication_count)
        parser.add_argument(
            '--rule',
            help='the name of rule')
        parser.add_argument(
            '--type',
            help='override the type of bucket shown in the report')
        parser.add_argument(
            '--choose-args',
            help='modify the weights')
        parser.add_argument(
            '--crushmap',
            help='path to the crushmap file')
        parser.add_argument(
            '-w', '--weights',
            help='path to the weights file')
        values_count = 100000
        parser.add_argument(
            '--values-count',
            help='repeat mapping (default: %d)' % values_count,
            type=int,
            default=values_count)
        return parser

    @staticmethod
    def set_parser(subparsers, arguments):
        parser = Analyze.get_parser()
        arguments(parser)
        subparsers.add_parser(
            'analyze',
            formatter_class=argparse.RawDescriptionHelpFormatter,
            description=textwrap.dedent("""\
            Analyze a crushmap rule

            The first step shows if the distribution run by a
            simulation is different from what is expected with the
            weights assigned to each item in crushmap.

            Map a number of objects (--values-count) to devices (three
            by default or --replication-count if specified) using a
            crush rule (--rule) from a given crushmap (--crushmap) and
            display a report comparing the expected and the actual
            object distribution.

            The format of the crushmap file specified with --crushmap
            can either be:

            - a JSON representation of a crushmap as documented in the
              Crush.parse_crushmap() method

            - a Ceph binary, text or JSON crushmap compatible with
              Luminuous and below

            The --type argument changes the item type displayed in the
            report. For instance --type device shows the individual
            OSDs and --type host shows the machines that contain
            them. If --type is not specified, it defaults to the
            "type" argument of the first "choose*" step of the rule
            selected by --rule.

            The first item in the report will be the first to become
            full. For instance if the report starts with:

                    ~id~  ~weight~  ~objects~  ~over/under filled %~
            ~name~
            g9       -22  2.29             85                  10.40

            it means that the bucket g9 with id -22 and weight 2.29
            will be the first bucket of its type to become full. The
            actual usage of the host will be 10.4% over the expected
            usage, i.e. if the g9 host is expected to be 70%
            full, it will actually be 80.40% full.

            The ~over/under filled %~ is the variation between the
            expected item usage and the actual item usage. If it is
            positive the item is overfilled, if it is negative the item
            is underfilled.

            The second step shows the worst case scenario if a bucket
            in the failure domain is removed from the crushmap. The
            failure domain is the type argument of the crush rule.
            For instance in:

                ["chooseleaf", "firstn", 0, "type", "host"]

            the failure domain is the host. If there are four hosts in
            the crushmap, named host1, host2, etc. a simulation will
            be run with a crushmap in which only host1 was
            removed. Another simulation will be run with a crushmap
            where host2 was removed etc. The result of all simulations
            are aggregated together.

            The worst case scenario for each item type is when the
            overfull percentage is higher. It is displayed as follows:

                     ~over filled %~
            ~type~
            device             25.55
            host               22.45

            If a host fail, the worst case scenario is that a device
            will be 25.55% overfull or a host will be 22.45% overfull.

            """),
            epilog=textwrap.dedent("""
            Examples:

            Display the first host that will become full.

            $ crush analyze --values-count 100 --rule data \\
                            --crushmap tests/sample-crushmap.json
                    ~id~  ~weight~  ~objects~  ~over/under filled %~
            ~name~
            host2     -4       1.0         70                    5.0
            host0     -2       1.0         65                   -2.5
            host1     -3       1.0         65                   -2.5

            Display the first device that will become full.

            $ crush analyze --values-count 100 --rule data \\
                            --type device \\
                            --crushmap tests/sample-crushmap.json
                     ~id~  ~weight~  ~objects~  ~over/under filled %~
            ~name~
            device0     0       1.0         28                  26.00
            device4     4       1.0         24                   8.00
            device5     5       2.0         46                   3.50
            device3     3       2.0         44                  -1.00
            device2     2       1.0         21                  -5.50
            device1     1       2.0         37                 -16.75
            """),
            help='Analyze crushmaps',
            parents=[parser],
        ).set_defaults(
            func=Analyze,
        )

    @staticmethod
    def collect_dataframe(crush, child):
        paths = crush.collect_paths([child], collections.OrderedDict())
        #
        # verify all paths have bucket types in the same order in the hierarchy
        # i.e. always rack->host->device and not host->rack->device sometimes
        #
        key2pos = {}
        pos2key = {}
        for path in paths:
            keys = list(path.keys())
            for i in range(len(keys)):
                key = keys[i]
                if key in key2pos:
                    assert key2pos[key] == i
                else:
                    key2pos[key] = i
                    pos2key[i] = key
        columns = []
        for pos in sorted(pos2key.keys()):
            columns.append(pos2key[pos])
        rows = []
        for path in paths:
            row = []
            for column in columns:
                element = path.get(column, np.nan)
                row.append(element)
                if element is not np.nan:
                    item_name = element
            item = crush.get_item_by_name(item_name)
            rows.append([item['id'],
                         item_name,
                         item.get('weight', 1.0),
                         item.get('type', 'device')] + row)
        d = pd.DataFrame(rows, columns=['~id~', '~name~', '~weight~', '~type~'] + columns)
        return d.set_index('~name~')

    @staticmethod
    def collect_cropped_weights(d, replication_count, failure_domain):
        d['~overweighted~'] = False
        d['~cropped weight~'] = d['~weight~'].copy()
        d['~cropped %~'] = 0.0
        for type in (failure_domain, 'device'):
            if len(d.loc[d['~type~'] == type]) == 0:
                continue
            w = d.loc[d['~type~'] == type].copy()
            tw = w['~weight~'].sum()
            w['~overweighted~'] = w['~weight~'].apply(lambda w: w > tw / replication_count)
            overweighted_count = len(w.loc[w['~overweighted~']])
            if overweighted_count > 0:
                tw_not_overweighted = w.loc[~w['~overweighted~'], ['~weight~']].sum()['~weight~']
                assert replication_count > overweighted_count
                cropped_weight = tw_not_overweighted / (replication_count - overweighted_count)
                w.loc[w['~overweighted~'], ['~cropped weight~']] = cropped_weight
                w['~cropped %~'] = (1.0 - w['~cropped weight~'] / w['~weight~']) * 100
            d.loc[d['~type~'] == type] = w
        return d

    @staticmethod
    def collect_nweight(d):
        d['~nweight~'] = 0.0
        for type in d['~type~'].unique():
            w = d.loc[d['~type~'] == type].copy()
            tw = w['~cropped weight~'].sum()
            w['~nweight~'] = w['~cropped weight~'].apply(lambda w: w / float(tw))
            d.loc[d['~type~'] == type] = w
        return d

    @staticmethod
    def collect_expected_objects(d, total):
        d['~expected~'] = 0
        for type in d['~type~'].unique():
            e = d.loc[d['~type~'] == type].copy()
            e['~expected~'] = e['~nweight~'].apply(lambda w: total * w).astype(int)
            remainder = total - e['~expected~'].sum()
            if remainder > 0:
                rounding = e['~expected~'].copy()
                rounding[:remainder] += 1
                e['~expected~'] = rounding
                assert total - e['~expected~'].sum() == 0
            d.loc[d['~type~'] == type] = e
        return d

    def collect_usage(self, d, total_objects):
        capacity = d['~nweight~'] * float(total_objects)
        n = self.main.value_name()
        d['~over/under filled %~'] = (d['~' + n + '~'] / capacity - 1.0) * 100 - d['~cropped %~']
        return d

    def run_simulation(self, c, root_name, failure_domain):
        if self.args.weights:
            with open(self.args.weights) as f_weights:
                weights = c.parse_weights_file(f_weights)
        else:
            weights = None

        values = self.main.hook_create_values()
        replication_count = self.args.replication_count
        total_objects = replication_count * len(values)

        root = c.find_bucket(root_name)
        log.debug("root = " + str(root))
        d = Analyze.collect_dataframe(c, root)
        d = Analyze.collect_cropped_weights(d, replication_count, failure_domain)
        d = Analyze.collect_nweight(d)
        d = Analyze.collect_expected_objects(d, total_objects)

        rule = self.args.rule
        device2count = collections.defaultdict(lambda: 0)
        for (name, value) in values.items():
            m = c.map(rule, value, replication_count, weights, choose_args=self.args.choose_args)
            if len(m) != replication_count:
                raise BadMapping("{} mapped to {}".format(value, m))
            for device in m:
                device2count[device] += 1

        item2path = c.collect_item2path([root])
        log.debug("item2path = " + str(item2path))
        d['~' + self.main.value_name() + '~'] = 0
        for (device, count) in device2count.items():
            for item in item2path[device]:
                d.at[item, '~' + self.main.value_name() + '~'] += count

        return self.collect_usage(d, total_objects)

    def analyze_failures(self, c, take, failure_domain):
        if failure_domain == 0:  # failure domain == device is a border case
            return None
        root = c.find_bucket(take)
        worst = pd.DataFrame()
        for may_fail in c.collect_buckets_by_type([root], failure_domain):
            f = Crush(verbose=self.args.debug,
                      backward_compatibility=self.args.backward_compatibility)
            f.crushmap = copy.deepcopy(c.get_crushmap())
            root = f.find_bucket(take)
            f.filter(lambda x: x.get('name') != may_fail.get('name'), root)
            f.parse(f.crushmap)
            try:
                a = self.run_simulation(f, take, failure_domain)
                a['~over filled %~'] = a['~over/under filled %~']
                a = a[['~type~', '~over filled %~']]
                worst = pd.concat([worst, a]).groupby(['~type~']).max().reset_index()
            except BadMapping:
                log.error("mapping failed when removing {}".format(may_fail))

        return worst.set_index('~type~')

    def _format_report(self, d, type):
        s = (d['~type~'] == type) & (d['~weight~'] > 0)
        n = self.main.value_name()
        a = d.loc[s, ['~id~', '~weight~', '~' + n + '~', '~over/under filled %~']]
        return str(a.sort_values(by='~over/under filled %~', ascending=False))

    def analyze_crushmap(self, crushmap):
        c = Crush(backward_compatibility=self.args.backward_compatibility)
        c.parse(crushmap)
        (take, failure_domain) = c.rule_get_take_failure_domain(self.args.rule)
        return self.run_simulation(c, take, failure_domain)

    def analyze(self):
        c = Crush(backward_compatibility=self.args.backward_compatibility)
        c.parse(self.main.convert_to_crushmap(self.args.crushmap))
        (take, failure_domain) = c.rule_get_take_failure_domain(self.args.rule)
        d = self.run_simulation(c, take, failure_domain)
        worst = self.analyze_failures(c, take, failure_domain)
        return (d, worst, failure_domain)

    def analyze_report(self, d, worst, failure_domain):
        if self.args.type:
            type = self.args.type
        else:
            type = failure_domain
        pd.set_option('precision', 2)
        out = ""
        out += self._format_report(d, type)
        if worst is not None:
            out += "\n\nWorst case scenario if a " + str(failure_domain) + " fails:\n\n"
            out += str(worst)
        if d['~overweighted~'].any():
            out += "\n\nThe following are overweighted and should be cropped:\n\n"
            out += str(d.loc[d['~overweighted~'],
                             ['~id~', '~weight~', '~cropped weight~', '~cropped %~']])
        return out

    def run(self):
        if not self.args.crushmap:
            raise Exception("missing --crushmap")
        return self.analyze_report(*self.analyze())



import os


def build_pre_hook(cmdobj):

    os.system(""" set -ex
    rm -fr build/tmp
    mkdir -p build/tmp
    ( cd build/tmp ; cmake ../../crush/libcrush )
    cp build/tmp/acconfig.h crush/libcrush
    rm -fr build/tmp
    """)


def build_post_hook(cmdobj):
    os.unlink("crush/libcrush/acconfig.h")



# -*- mode: python; coding: utf-8 -*-
#
# Copyright (C) 2017 <contact@redhat.com>
#
# Author: Loic Dachary <loic@dachary.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
import logging
import pytest # noqa needed for capsys

from crush import Crush
from crush.main import Main
from crush.analyze import Analyze

logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s',
                    level=logging.DEBUG)


class TestAnalyze(object):

    def test_collect_dataframe(self):
        tree = {
            'name': 'rack0', 'type': 'rack', 'id': -1, 'children': [
                {'name': 'host0', 'type': 'host', 'id': -2, 'children': [
                    {'name': 'osd.3', 'id': 3},
                ]},
                {'name': 'host1', 'type': 'host', 'id': -3, 'children': [
                    {'name': 'osd.4', 'id': 4},
                ]},
            ]
        }
        c = Crush(verbose=1)
        c.parse({"trees": [tree]})
        d = Analyze.collect_dataframe(c, tree)
        expected = """\
        ~id~  ~weight~  ~type~   rack   host device
~name~                                             
rack0     -1       1.0    rack  rack0    NaN    NaN
host0     -2       1.0    host  rack0  host0    NaN
osd.3      3       1.0  device  rack0  host0  osd.3
host1     -3       1.0    host  rack0  host1    NaN
osd.4      4       1.0  device  rack0  host1  osd.4\
""" # noqa trailing whitespaces are expected
        assert expected == str(d)

    def make_analyze(self, size, host_weight):
        p = [
            '--replication-count', str(size),
            '--values-count', '2048',
        ]

        hosts_count = len(host_weight)
        crushmap = {
            "trees": [
                {
                    "type": "root",
                    "id": -1,
                    "name": "dc1",
                    "weight": sum(host_weight),
                    "children": [],
                }
            ],
            "rules": {
                "firstn": [
                    ["take", "dc1"],
                    ["set_choose_tries", 100],
                    ["choose", "firstn", 0, "type", "host"],
                    ["emit"]
                ],
            }
        }
        crushmap['trees'][0]['children'].extend([
            {
                "type": "host",
                "id": -(i + 2),
                "name": "host%d" % i,
                "weight": host_weight[i],
                "children": [],
            } for i in range(0, hosts_count)
        ])
        a = Main().constructor([
            'analyze',
            '--rule', 'firstn',
        ] + p)
        a.args.crushmap = crushmap
        return a

    def test_analyze_out_of_bounds(self):

        weights = [7, 7, 7, 3, 3]
        print("==== weights " + str(weights))
        a = self.make_analyze(4, weights)
        d = a.analyze_report(*a.analyze())
        print(str(d))
        expected = """\
        ~id~  ~weight~  ~objects~  ~over/under filled %~
~name~                                                  
host4     -6         3       1329                  29.79
host3     -5         3       1325                  29.39
host1     -3         7       1857                 -23.61
host0     -2         7       1848                 -24.05
host2     -4         7       1833                 -24.78

Worst case scenario if a host fails:

        ~over filled %~
~type~                 
host                0.0
root                0.0

The following are overweighted and should be cropped:

        ~id~  ~weight~  ~cropped weight~  ~cropped %~
~name~                                               
host0     -2         7               6.0        14.29
host1     -3         7               6.0        14.29
host2     -4         7               6.0        14.29\
""" # noqa trailing whitespaces are expected
        assert expected == str(d)

        weights = [5, 1, 1, 1, 1]
        print("==== weights " + str(weights))
        a = self.make_analyze(2, weights)
        d = a.analyze_report(*a.analyze())
        print(str(d))

        expected = """\
        ~id~  ~weight~  ~objects~  ~over/under filled %~
~name~                                                  
host4     -6         1        617                  20.51
host3     -5         1        612                  19.53
host2     -4         1        593                  15.82
host1     -3         1        584                  14.06
host0     -2         5       1690                 -37.48

Worst case scenario if a host fails:

        ~over filled %~
~type~                 
host              17.19
root               0.00

The following are overweighted and should be cropped:

        ~id~  ~weight~  ~cropped weight~  ~cropped %~
~name~                                               
host0     -2         5               4.0         20.0\
""" # noqa trailing whitespaces are expected
        assert expected == str(d)

        weights = [7, 7, 3, 1, 1, 1]
        print("==== weights " + str(weights))
        a = self.make_analyze(3, weights)
        d = a.analyze_report(*a.analyze())
        print(str(d))
        expected = """\
        ~id~  ~weight~  ~objects~  ~over/under filled %~
~name~                                                  
host3     -5         1        468                  37.11
host4     -6         1        461                  35.06
host5     -7         1        451                  32.13
host2     -4         3       1215                  18.65
host0     -2         7       1791                 -26.83
host1     -3         7       1758                 -28.45

Worst case scenario if a host fails:

        ~over filled %~
~type~                 
host              30.86
root               0.00

The following are overweighted and should be cropped:

        ~id~  ~weight~  ~cropped weight~  ~cropped %~
~name~                                               
host0     -2         7               6.0        14.29
host1     -3         7               6.0        14.29\
""" # noqa trailing whitespaces are expected
        assert expected == str(d)

        weights = [5, 5, 3, 3, 3]
        print("==== weights " + str(weights))
        a = self.make_analyze(4, weights)
        d = a.analyze_report(*a.analyze())
        print(str(d))
        expected = """\
        ~id~  ~weight~  ~objects~  ~over/under filled %~
~name~                                                  
host3     -5         3       1535                  12.43
host4     -6         3       1504                  10.16
host2     -4         3       1484                   8.69
host0     -2         5       1835                 -20.40
host1     -3         5       1834                 -20.45

Worst case scenario if a host fails:

        ~over filled %~
~type~                 
host                0.0
root                0.0

The following are overweighted and should be cropped:

        ~id~  ~weight~  ~cropped weight~  ~cropped %~
~name~                                               
host0     -2         5               4.5         10.0
host1     -3         5               4.5         10.0\
""" # noqa trailing whitespaces are expected
        assert expected == str(d)

    def test_analyze(self):
        trees = [
            {"name": "dc1", "type": "root", "id": -1, 'children': []},
        ]
        weights = (
            (100, 10, 50, 40),
            (100, 10, 50, 40),
            (100, 10, 50, 40),
            (100, 10, 50, 40),
            (10, 1, 5, 4),
        )
        trees[0]['children'].extend([
            {
                "type": "host",
                "id": -(i + 3),
                "name": "host%d" % i,
                "weight": weights[i][0],
                "children": [
                    {"id": (3 * i),
                     "name": "device%02d" % (3 * i), "weight": weights[i][1]},
                    {"id": (3 * i + 1),
                     "name": "device%02d" % (3 * i + 1), "weight": weights[i][2]},
                    {"id": (3 * i + 2),
                     "name": "device%02d" % (3 * i + 2), "weight": weights[i][3]},
                ],
            } for i in range(5)
        ])
        a = Main().constructor([
            'analyze',
            '--rule', 'data',
            '--replication-count', '2',
            '--values-count', '10000',
        ])
        a.args.crushmap = {
            "trees": trees,
            "rules": {
                "data": [
                    ["take", "dc1"],
                    ["chooseleaf", "firstn", 0, "type", "host"],
                    ["emit"]
                ]
            }
        }
        d = a.analyze_report(*a.analyze())
        print(str(d))
        expected = """\
        ~id~  ~weight~  ~objects~  ~over/under filled %~
~name~                                                  
host4     -7      10.0        541                  10.91
host3     -6     100.0       4930                   1.07
host2     -5     100.0       4860                  -0.37
host1     -4     100.0       4836                  -0.86
host0     -3     100.0       4833                  -0.92

Worst case scenario if a host fails:

        ~over filled %~
~type~                 
device            25.55
host              22.45
root               0.00\
""" # noqa trailing whitespaces are expected
        assert expected == str(d)

    def test_analyze_weights(self):
        a = Main().constructor(
            ["analyze", "--rule", "replicated_ruleset",
             "--replication-count", "2", "--type", "device",
             "--crushmap", "tests/weights-crushmap.json",
             "--weights", "tests/weights.json"])
        a.args.backward_compatibility = True
        res = a.run()
        assert "-100.00" in str(res)  # One of the OSDs has a weight of 0.0

# Local Variables:
# compile-command: "cd .. ; tox -e py27 -- -s -vv tests/test_analyze.py"
# End:



# -*- mode: python; coding: utf-8 -*-
#
# Copyright (C) 2017 <contact@redhat.com>
#
# Author: Loic Dachary <loic@dachary.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import division

import collections
import copy
import json
import logging
from crush.libcrush import LibCrush

log = logging.getLogger(__name__)


class Crush(object):
    """Control object placement in a hierarchy.

    The algorithms optimize the placement so that:

    - the devices are filled according to their weight

    - the number of objects that move is proportional to the magnitude
      of the weight of the devices removed or added

    **Choose and chooseleaf retries**

    The "choose firstn" and "choose indep" rule step look for buckets
    of a given type, randomly selecting them. If they are unlucky and
    find the same bucket twice, they will try N+1 times (N being the
    value of the choose_total_tries tunable). If there is a previous
    "set_choose_tries" step in the same rule, it will try O times
    instead (O being the value of the argument of the
    "set_choose_tries" step).

    Note: the choose_total_tries tunable is the number of retry, not
    the number of tries. The number of tries is the number of retry +
    1. The "set_choose_tries" rule step sets the number of tries and
    does not need the + 1. This confusing difference is inherited from
    an off-by-one bug from years ago.

    The "chooseleaf firstn" and "chooseleaf indep" rule step do the same
    as "choose firstn" and "choose indep" but also recursively explore
    each bucket found, looking for a single device. The same device
    may be found in two different buckets because the crush map is not
    a strict hierarchy, it is a DAG. When such a collision happens,
    they will try again. The number of times they try to find a non
    colliding device is:

    - If "firstn" and there is no previous "set_chooseleaf_tries" rule step: try
      N + 1 times (N being the value of the choose_total_tries tunable)

    - If "firstn" and there is a previous "set_chooseleaf_tries" rule
      step: try P times (P being the value of the argument of the
      "set_chooseleaf_tries" rule step)

    - If "indep" and there is no previous "set_chooseleaf_tries" rule step: try
      1 time.

    - If "indep" and there is a previous "set_chooseleaf_tries" rule step:
      try P times (P being the value of the argument of the
      "set_chooseleaf_tries" rule step)

    **Backward compatibility**

    Some tunables and rule steps only exist for backward
    compatibility. Trying to use them will raise an exception, unless
    Crush is created with backward_compatibility=True. They are listed
    here for reference but they are documented in the libcrush api
    at http://doc.libcrush.org/master/group___a_p_i.html.

    In the crushmap::

            tunables = {
              "choose_local_tries": 0,
              "choose_local_fallback_tries": 0,
              "chooseleaf_vary_r": 1,
              "chooseleaf_stable": 1,
              "chooseleaf_descend_once": 1,
              "straw_calc_version": 1,
            }

    In a rule step::

            [ "set_choose_local_tries", <positive integer> ]
            [ "set_choose_local_fallback_tries", <positive integer> ]
            [ "set_chooseleaf_vary_r", 0 or 1 ]
            [ "set_chooseleaf_stable", 0 or 1 ]

    """

    def __init__(self, verbose=False, backward_compatibility=False):
        """Create a Crush.

        If the optional argument `verbose` is set to True, all methods
        will print debug information on stdout. It defaults to False.

        If the optional argument `backward_compatibility` is set to
        True, the tunables designed for backward_compatibility are
        allowed, otherwise trying to use them raises an exception. See
        **Backward compatibility** in the class documentation.

        """
        self.c = LibCrush(verbose=verbose and 1 or 0,
                          backward_compatibility=backward_compatibility and 1 or 0)

    def parse(self, something):
        """Validate and parse `something` which can be one of the
        following:

        - a crushmap dict as documented in the parse_crushmap() method

        - a path to a file containing a JSON representation of a
          crushmap as documented in the Crush.parse_crushmap() method

        - a path to a file containing a Ceph binary, text or JSON
          crushmap compatible with Luminuous and below

        The details of the validation and parsing are documented
        in the parse_crushmap() method.

        """
        return self.parse_crushmap(self._convert_to_crushmap(something))

    def parse_crushmap(self, crushmap):
        """Validate and parse the `crushmap` object.

        The `crushmap` is a hierarchical description of devices in
        which objects can be stored and rules to place the objects.
        It is verified to obey the specifications below. An exception
        is raised on the first error found.
        ::

            crushmap = {
              # optional (default: none)
              "trees": trees,

              # optional (default: none)
              "rules": rules,

              # optional (default: none)
              "choose_args": choose_arg_map,

              # optional (default: none)
              "tunables": tunables,
            }

        The "trees" are the roots of device hierarchies, the "rules" describe
        various object placement strategies for this device hierarchy.
        ::

            trees = [
              # optional (default: none)
              bucket or device or reference,
              bucket or device or reference,
              ...
            ]

        Each element can either be a device (i.e. no children and id >= 0),
        a bucket (i.e. id < 0) or a reference to a bucket or device.
        ::

            bucket = {
              # mandatory
              "type": <str>,

              # mandatory
              "name": <str>,

              # optional (default: first available id)
              "id": <negative int>,

              # optional (default: "straw2")
              "algorithm": "uniform" or "list" or "straw2",

              # optional (default: cumulated children weights or none)
              "weight": <postive Q16.16>,

              # optional (default: none)
              "children": children
            }

        The **type** is a user defined string that can be used by
        **rules** to select all buckets of the same type.

        The **name** is a user defined string that uniquely identifies
        the bucket.

        The **id** must either be set for all buckets or not at
        all. If the **id** is provided, it must be a unique negative
        number. If it is not provided, the first available id is
        used.

        The **weight** must either be set for all buckets or not at
        all. If not set, **weight** defaults to the cumulated weight
        of the immediate children bucket or devices or reference,
        recursively, bottom to top.

        Children within a bucket are chosen with one of three
        **algorithms** representing a tradeoff between performance
        and reorganization efficiency. If you are unsure, we recommend
        using **"straw2"**. The table summarizes how the speed of each
        option measures up against mapping stability when items are
        added or removed::

            Bucket Alg     Speed       Additions    Removals
            ------------------------------------------------
            "straw2"        O(n)       optimal      optimal
            "uniform"       O(1)       poor         poor
            "list"          O(n)       optimal      poor

        - **"straw2"**: List and tree buckets are structured such that
          a limited number of hash values need to be calculated and
          compared to weights in order to select a bucket child. In
          doing so, they divide and conquer in a way that either gives
          certain children precedence (e. g., those at the beginning
          of a list) or obviates the need to consider entire subtrees
          of children at all. That improves the performance of the
          replica placement process, but can also introduce suboptimal
          reorganization behavior when the contents of a bucket change
          due an addition, removal, or re-weighting of an child. The
          straw2 bucket type allows all children to fairly `compete`
          against each other for replica placement through a process
          analogous to a draw of straws. To place a replica, a straw
          of random length is drawn for each child in the bucket.  The
          child with the longest straw wins.  The length of each straw
          is initially a value in a fixed range.  Each straw length is
          scaled by a factor based on the child’s weight so that
          heavily weighted children are more likely to win the draw.
          Although this process is almost twice as slow (on average)
          than a list bucket and even slower than a tree bucket (which
          scales logarithmically), straw2 buckets result in optimal
          data movement between nested children when modified.

        - **"uniform"**: Devices are rarely added individually in a
          large system.  Instead, new storage is typically deployed in
          blocks of identical devices, often as an additional shelf in
          a server rack or perhaps an entire cabinet. Devices reaching
          their end of life are often similarly decommissioned as a
          set (individual failures aside), making it natural to treat
          them as a unit. Uniform buckets are used to represent an
          identical set of devices in such circumstances. The key
          advantage in doing so is performance related: Crush can map
          replicas into uniform buckets in constant time. In cases
          where the uniformity restrictions are not appropriate, other
          bucket types can be used. If the size of a uniform bucket
          changes, there is a complete reshuffling of data between
          devices, much like conventional hash-based distribution
          strategies.

        - **"list"**: List buckets structure their contents as a
          linked list, and can contain children with arbitrary
          weights.  To place a replica, Crush begins at the head of
          the list with the most recently added child and compares its
          weight to the sum of all remaining children’ weights.
          Depending on the value of the hash function, either the
          current child is chosen with the appropriate probability, or
          the process continues recursively down the list. This is a
          natural and intuitive choice for an expanding cluster:
          either an object is relocated to the newest device with some
          appropriate probability, or it remains on the older devices
          as before.  The result is optimal data migration when
          children are added to the bucket. Children removed from the
          middle or tail of the list, however, can result in a
          significant amount of unnecessary movement, making list
          buckets most suitable for circumstances in which they never
          (or very rarely) shrink.

        There also exists an older version of "straw2", known as **"straw"**.
        Given the choice, it is *always* better to use "straw2"; "straw" is
        therefore not allowed by default. If the crushmap being read is old
        and still uses "straw", setting `backward_compatibility=True` when
        creating the `Crush` object will allow it to be used.
        ::

            children = [
              # optional (default: none)
              bucket or device or reference,
              bucket or device or reference,
              ...
            ]

        Each element can either be a device (i.e. no children and id >= 0),
        a bucket (i.e. id < 0) or a reference to a bucket or device.
        ::

            reference = {
              # mandatory
              "reference_id": <int>,

              # optional (default: 1.0)
              "weight": <postive Q16.16>,
            }

        The **reference_id** must be equal to the **id** of a bucket or
        device defined in the hierarchy.

        If the **weight** is omitted, it default to 0x10000. The
        **weight** must either be set for all references or not at
        all.
        ::

            device = {
              # mandatory
              "id": <positive int>,

              # mandatory
              "name": <str>,

              # optional (default: 0x10000)
              "weight": <postive Q16.16>,
            }

        The **id** must be a unique positive number.

        The **name** is a user defined string that uniquely identifies
        the device.

        If the **weight** of a device A is lower than the
        **weight** of a device B, it will be less likely to be used.
        A common pattern is to set the **weight** to 0x20000 for 2TB
        devices, 0x10000 for 1TB devices, 0x08000 for 500GB devices, etc.
        ::

            rules = {
              # optional (default: none)
              <rule name str>: rule,
              <rule name str>: rule,
              ...
            }

        A **rule** maps an object (see the map function) to a list
        of devices. There can be multiple rules depending on the
        mapping strategy.
        ::

            rule = [ step, step, ... ]

        The **rule** interprets each **step** in sequence and the last
        one must be **"emit"**.
        ::

            step = [ "take", <bucket name str> ]

        Select the **bucket name**.
        ::

            step = [
              "set_choose_tries" or "set_chooseleaf_tries",
              <positive integer>
            ]

        Overrides the default number of tries (set by the
        choose_total_tries tunables and defaulting to 50) when looking
        for a bucket (set_choose_tries) or a device
        (set_chooseleaf_tries). See **Choose and chooseleaf retries**
        in the class documentation for more information.
        ::

            step = [
              "choose"
              "firstn" or "indep",
              <replication count positive int>,
              "type"
              bucket type or device type
            ]

        Recursively explore each bucket currently selected, looking for
        **replication count** buckets of the required **bucket type**
        and select them.

        **firstn** is best suited when the order of the result does not matter,
        for instance to place object replicas. **indep** is best suited when
        the order of the result makes a difference, for instance to place parts
        of an erasure coded object.

        If **replication count** is zero, the number of buckets
        to select will be determined by the `replication_count` argument of
        the `map` method, i.e. **replication count** is set to
        match the desired number of replicas.
        ::

            bucket type = <str>

        The type field of a bucket definition is used in rule steps to
        designate all buckets with the same type.
        ::

            device type = 0

        The devices all have the same fixed type: 0.
        ::

            step = [
              "chooseleaf"
              "firstn" or "indep",
              <replication count positive int>,
              "type"
              <bucket type str>
            ]

        Recursively explore each bucket currently selected, looking for
        **replication count** devices within all buckets of
        the required **bucket type** and select them.

        **firstn** is best suited when the order of the result does not matter,
        for instance to place object replicas. **indep** is best suited when
        the order of the result makes a difference, for instance to place parts
        of an erasure coded object.

        If **replication count** is zero, the number of devices
        to select will be determined by the `replication_count` argument of
        the `map` method, i.e. **replication count** is set to
        match the desired number of replicas.
        ::

            step = [ "emit" ]

        Append the selection to the results and clear the selection
        ::

            choose_arg_map = {
              <name str>: choose_args,
              <name str>: choose_args,
              ...
            }

        A named collection of **choose_args** for **map()** to use for
        changing the weights of straw2 buckets or remap the items they
        contains to arbitrary values. See **map()** for more information.
        ::

            choose_args = [
              choose_args_bucket,
              choose_args_bucket,
              ...
            ]

        A list of **choose_args_bucket**, each modifying the weights
        or the ids of a **bucket**. There must be at most one
        **choose_args_bucket** for a given bucket in the crushmap.
        ::

            choose_args_bucket = {
              # mandatory
              "bucket_id": <negative int> or "bucket_name": <bucket name str>,
              # optional (default: none)
              "ids": choose_args_ids,
              # optional (default: none)
              "weight_set": choose_args_weight_set,
            }

        When calling **map()** with this **choose_args_bucket**,
        straw2 will use **choose_args_ids** instead of the items it
        contains. And instead of using the weights stored in the
        bucket, it will use one of **choose_args_weight_set**,
        depending on the position. See **map()** for more information.

        The bucket is uniquely identified by either **bucket_id**
        which must be a negative number or **bucket_name** which is
        the name of the bucket set in the **bucket** definition. The
        two are mutually exclusive.
        ::

            choose_args_ids = [ <int>, <int>, ... ]

        The first element of **choose_args_ids** will be used instead
        of the id of the first children and so on.

        The size of **choose_args_ids** must be exactly the same as
        the size of the **children** array of the corresponding
        **bucket**.
        ::

            choose_args_weight_set = [
               [ <positive Q16.16>, <positive Q16.16>, ... ], # position 0
               [ <positive Q16.16>, <positive Q16.16>, ... ], # position 1
               ...
            ]

        When **map()** chooses the frist replicas from the
        corresponding **bucket** it will use the weights at position 0
        instead of the weights stored in the **bucket**. When choosing
        the second replica it will use the weights at positon 1 and so
        on.

        The size of the array for each position must be exactly the
        same as the size of the **children** array of the corresponding
        **bucket**.

        If **choose_args_weight_set** does not contain a list of weights
        for a given position, the weights in the last available position
        will be used instead.
        ::

            tunables = {
              "choose_total_tries": 50
            }

        A tunable changes the behavior of the **map()** method and it
        will not return the same list of devices if it is changed.

        - **choose_total_tries** (optional positive integer, default:
          50) is the default number of retries on a collision.
          See **Choose and chooseleaf retries** in the class
          documentation for more information.

        """
        self.crushmap = copy.deepcopy(crushmap)
        self.c.parse(self.crushmap)
        self._update_info()
        return True

    def map(self, rule, value, replication_count, weights=None, choose_args=None):
        """Map an object to a list of devices.

        The **rule** is used to map the **value** (representing an
        object) to the desired number of devices
        (**replication_count**) and return them in a list. The
        probabilities for a given device to be selected can be
        modified by the **weights** dictionnary or by **choose_args**
        for straw2 buckets.

        If the mapping is successful, a list of device containing
        exactly **replication_count** devices is returned. If the
        mapping fails, the list may contains less devices or some
        names may be replaced by None. For instance, if asking
        for 3 replicas, the result of a failed mapping could be::

            [ "device1", "device5" ] # 2 instead of 3
            [ "device8", None, "device0" ] # second device is missing

        The dictionary of **weights** modifies the probabilities for
        the device names it contains. The value is a float in the
        range [0..1]. If the weight is 0, the device will never be
        selected.  If the weight is 1, the probability that the device
        is selected is not modified. If a device is not in the
        **weights** dictionary, its probability is not modified. For
        example::

            { "device0": 0.50, "device1": 0.75 }

        will multiply the probability of "device0" by 0.50 (i.e,
        reduce it by 50%) and "device1" by 0.75 (i.e, reduce it by
        25%).

        If the **choose_args** is a string, the corresponding list
        will be retrieved from **choose_arg_map** in the map. Each
        element in the **choose_args** list modifies the parameters of
        the choose function for the corresponding straw2 bucket. With
        a bucket containing two children::

            {
               "name": "bucket1":
               "children": [
                   { "name": "device1", "id": 1, "weight": 1 * 0x10000 },
                   { "name": "device2", "id": 2, "weight": 2 * 0x10000 },
               ]
            }

        If map(85) == [ 'device2' ], the result could be different
        by switching the weights with::

            map(85, choose_args=[
              { "bucket_name": "bucket1",
                "weight_set": [ [ 2 * 0x10000, 1 * 0x10000 ] ]
              }]) == [ 'device1' ]

        Similarly the result can be influenced by providing
        alternative ids. The id of each item is a parameter of the
        hash considered for placement. However, contrary to the
        weight, there is no way to guess how it will influence the
        result.::

            map(85, choose_args=[
              { "bucket_name": "bucket1",
                "ids": [ 100, 300 ]
              }]) == [ 'device1' ]

        - **rule**: the rule name (required string)

        - **value**: the number to map (required integer)

        - **replication_count**: the desired number of devices
            (required positive integer)

        - **weights**: map of name to weight float (optional, default to None)

        - **choose_args**: name to lookup in the map or a list (optional, default to None)

        Return a list of device names.

        """
        kwargs = {
            "rule": rule,
            "value": value,
            "replication_count": replication_count,
        }
        if weights:
            kwargs["weights"] = weights
        if choose_args:
            kwargs["choose_args"] = choose_args
        return self.c.map(**kwargs)

    def _convert_to_crushmap(self, something):
        if type(something) in (dict, collections.OrderedDict):
            return something
        else:
            with open(something) as f_json:
                return json.load(f_json, object_pairs_hook=collections.OrderedDict)

    def to_file(self, out_path):
        open(out_path, "w").write(json.dumps(self.crushmap, indent=4, sort_keys=True))

    @staticmethod
    def parse_osdmap_weights(osdmap):
        weights = {}
        for osd in osdmap["osds"]:
            if osd["weight"] < 1.0:
                key = "osd.{}".format(osd["osd"])
                weights[key] = osd["weight"]
        return weights

    @staticmethod
    def parse_weights_file(weights_file):
        """
        Parse a file containing information about devices weights.
        The file is expected to contain JSON data. It can either be:

         - a dictionary that maps weights (as floats between 0 and 1)
           to device names, e.g.::

            {"device0": 0.71, "device1": 0.212}

         - a JSON dump of a Ceph OSDMap, obtainable with the command::

            ceph osd dump --format json > osdmap.json

        :param weights_file: File object to the weights file
        :type weights_file: file
        :return: Parsed weights
        :rtype: dict[str, float]
        """

        f_name = weights_file.name

        # Load the JSON data
        try:
            raw_data = json.load(weights_file)
        except ValueError:
            raise AssertionError("{} is not valid JSON".format(f_name))

        # It should be an object (dict) and not an array (list)
        assert type(raw_data) is dict, "Expected {} to be a dict".format(f_name)

        osdmap_keys = {"epoch", "fsid", "osds", "pools"}  # And many more!
        if all(k in raw_data for k in osdmap_keys):
            weights = Crush.parse_osdmap_weights(raw_data)
        else:
            weights = raw_data

        # Check the weight values
        assert all(type(v) is float and 0 <= v <= 1 for v in weights.values()), \
            "Weights must be floating-point values between 0 and 1"

        # Don't check that the keys are existing devices, LibCrush will do it
        # No need to check that keys are strings, it's enforced by JSON

        return weights

    #
    # Working with the crushmap in memory structure
    #
    def get_crushmap(self):
        """
        Return the original crushmap given to the parse() method.

        The returned crushmap does not contain any reference_id,
        they are replaced by a pointer to the actual bucket. This
        is convenient when exploring the crushmap. But it will
        fail to parse again because duplicated buckets will be
        found.
        """
        return self.crushmap

    def _collect_items(self, children):
        for child in children:
            if 'id' in child:
                self._name2item[child['name']] = child
                self._id2item[child['id']] = child
            self._collect_items(child.get('children', []))

    def _update_info(self):
        self._name2item = {}
        self._id2item = {}
        trees = self.crushmap.get('trees', [])
        self._collect_items(trees)

    def get_item_by_id(self, id):
        return self._id2item[id]

    def get_item_by_name(self, name):
        return self._name2item[name]

    def rule_get_take_failure_domain(self, name):
        rule = self.crushmap['rules'][name]
        take = None
        failure_domain = None
        for step in rule:
            if step[0] == 'take':
                assert take is None
                take = step[1]
            elif step[0].startswith('choose'):
                assert failure_domain is None
                (op, firstn_or_indep, num, _, failure_domain) = step
        return (take, failure_domain)

    def find_bucket(self, name):
        def walk(children):
            for child in children:
                if child.get('name') == name:
                    return child
                found = walk(child.get('children', []))
                if found:
                    return found
            return None
        return walk(self.crushmap.get('trees', []))

    @staticmethod
    def collect_buckets_by_type(root, type):
        def walk(children):
            found = []
            for child in children:
                if child.get('type') == type:
                    found.append(child)
                found.extend(Crush.collect_buckets_by_type(child.get('children', []), type))
            return found
        return walk(root)

    def _merge_choose_args(self):
        if 'choose_args' not in self.crushmap:
            return False
        id2choose_args = collections.defaultdict(lambda: {})
        for name, choose_args in self.crushmap['choose_args'].items():
            for choose_arg in choose_args:
                id2choose_args[choose_arg['bucket_id']][name] = choose_arg
        del self.crushmap['choose_args']

        def walk(bucket):
            if bucket['id'] in id2choose_args:
                bucket['choose_args'] = id2choose_args[bucket['id']]
            for child in bucket.get('children', []):
                walk(child)
        for root in self.crushmap['trees']:
            walk(root)
        return True

    def _split_choose_args(self):
        assert 'choose_args' not in self.crushmap
        name2choose_args = collections.defaultdict(lambda: [])

        def walk(bucket):
            if 'choose_args' in bucket:
                for name, choose_arg in bucket['choose_args'].items():
                    assert bucket['id'] == choose_arg['bucket_id']
                    name2choose_args[name].append(choose_arg)
                del bucket['choose_args']
            for child in bucket.get('children', []):
                walk(child)
        for root in self.crushmap['trees']:
            walk(root)
        self.crushmap['choose_args'] = {}
        for name, choose_args in name2choose_args.items():
            self.crushmap['choose_args'][name] = sorted(choose_args, key=lambda v: v['bucket_id'])

    def update_choose_args(self, name, choose_args):
        if 'choose_args' not in self.crushmap:
            self.crushmap['choose_args'] = {name: choose_args}
            return
        if name not in self.crushmap['choose_args']:
            self.crushmap['choose_args'][name] = choose_args
            return
        u = {}
        for choose_arg in self.crushmap['choose_args'][name] + choose_args:
            u[choose_arg['bucket_id']] = choose_arg
        self.crushmap['choose_args'][name] = sorted(u.values(), key=lambda v: v['bucket_id'])

    def filter(self, fun, root):
        names = self.crushmap.get('choose_args', {}).keys()
        self._merge_choose_args()

        def walk(bucket):
            for pos in reversed(range(len(bucket.get('children', [])))):
                if not fun(bucket['children'][pos]):
                    child = bucket['children'][pos]
                    del bucket['children'][pos]
                    if 'choose_args' in bucket:
                        for name, choose_arg in bucket['choose_args'].items():
                            if 'ids' in choose_arg:
                                del choose_arg['ids'][pos]
                            if 'weight_set' in choose_arg:
                                for weights in choose_arg['weight_set']:
                                    del weights[pos]
            for child in bucket.get('children', []):
                walk(child)
        walk(root)

        self._split_choose_args()
        if 'choose_args' in self.crushmap:
            for name in names:
                if name not in self.crushmap['choose_args']:
                    self.crushmap['choose_args'][name] = []

    @staticmethod
    def collect_paths(children, path):
        children_info = []
        for child in children:
            child_path = copy.copy(path)
            child_path[child.get('type', 'device')] = child['name']
            children_info.append(child_path)
            if child.get('children'):
                children_info.extend(Crush.collect_paths(child['children'], child_path))
        return children_info

    def collect_item2path(self, children):
        paths = self.collect_paths(children, collections.OrderedDict())
        item2path = {}
        for path in paths:
            elements = list(path.values())
            item2path[elements[-1]] = elements
        return item2path



# -*- mode: python; coding: utf-8 -*-
#
# Copyright (C) 2017 <contact@redhat.com>
#
# Author: Loic Dachary <loic@dachary.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
import pytest

from crush.libcrush import LibCrush

STEP_BACKWARDS = [
    "choose_local_tries",
    "choose_local_fallback_tries",
    "chooseleaf_vary_r",
    "chooseleaf_stable",
]
PARSE_BACKWARDS = STEP_BACKWARDS + [
    "chooseleaf_descend_once",
    "straw_calc_version",
]


class TestLibCrush(object):

    def test_parse_types_invalid(self):
        wrong = {
            'types': 1
        }
        with pytest.raises(RuntimeError) as e:
            LibCrush().parse(wrong)
        assert 'must be a list' in str(e.value)

        wrong = {
            'types': [
                1
            ]
        }
        with pytest.raises(RuntimeError) as e:
            LibCrush().parse(wrong)
        assert 'must be a dict' in str(e.value)

        wrong = {
            'types': [
                {}
            ]
        }
        with pytest.raises(RuntimeError) as e:
            LibCrush().parse(wrong)
        assert 'missing type_id' in str(e.value)

        wrong = {
            'types': [
                {
                    'type_id': []
                }
            ]
        }
        with pytest.raises(TypeError):
            LibCrush().parse(wrong)

        wrong = {
            'types': [
                {
                    'type_id': 1
                }
            ]
        }
        with pytest.raises(RuntimeError) as e:
            LibCrush().parse(wrong)
        assert 'missing name' in str(e.value)

        wrong = {
            'types': [
                {
                    'type_id': 1,
                    'name': []
                }
            ]
        }
        with pytest.raises(TypeError):
            LibCrush().parse(wrong)

    def test_parse_tunables(self, capsys):
        total_tries = 1234
        crushmap = {
            'tunables': {
                'choose_total_tries': total_tries,
            }
        }
        assert LibCrush(verbose=1).parse(crushmap)
        out, err = capsys.readouterr()
        assert 'choose_total_tries = ' + str(total_tries) in out

    def test_parse_tunables_invalid(self):
        wrong = {
            'tunables': 1
        }
        with pytest.raises(RuntimeError) as e:
            LibCrush().parse(wrong)
        assert 'must be a dict' in str(e.value)

        wrong = {
            'tunables': {
                1: 0
            }
        }
        with pytest.raises(TypeError):
            LibCrush().parse(wrong)

        wrong = {
            'tunables': {
                'choose_total_tries': 'wrong'
            }
        }
        with pytest.raises(TypeError):
            LibCrush().parse(wrong)

    def test_parse_tunables_backward_compatibity(self, capsys):
        for backward in PARSE_BACKWARDS:
            crushmap = {
                'tunables': {
                    backward: 1234,
                }
            }
            with pytest.raises(RuntimeError) as e:
                LibCrush(verbose=1).parse(crushmap)
            assert 'not allowed unless backward_compatibility' in str(e.value)
            LibCrush(verbose=1, backward_compatibility=1).parse(crushmap)

    def test_parse_verbose(self, capsys):
        empty = {
            'trees': [],
        }
        assert LibCrush(verbose=1).parse(empty)
        out, err = capsys.readouterr()
        assert 'trees' in out

        assert LibCrush().parse(empty)
        out, err = capsys.readouterr()
        assert 'trees' not in out

    def test_parse_empty(self):
        LibCrush().parse({})

    def test_parse_no_argument(self):
        with pytest.raises(TypeError):
            LibCrush().parse()

    def test_parse_argument_wrong(self):
        with pytest.raises(TypeError):
            LibCrush().parse([])

    def test_parse_trees_type_wrong(self):
        wrong = {
            'trees': {}
        }
        with pytest.raises(RuntimeError) as e:
            LibCrush().parse(wrong)
        assert 'must be a list' in str(e.value)

    def test_parse_invalid_algorithm(self):
        wrong = {
            'trees': [{
                'type': 'root',
                'algorithm': 'FOOBAR',
            }]
        }
        with pytest.raises(RuntimeError) as e:
            LibCrush().parse(wrong)
        assert 'not FOOBAR' in str(e.value)
        wrong = {
            'trees': [{
                'type': 'root',
                'algorithm': 0
            }]
        }
        with pytest.raises(TypeError) as e:
            LibCrush().parse(wrong)

    def test_parse_straw_algorithm(self):
        """Test that LibCrush only parses the straw algorithm
        if backward_compatibility is set to True"""
        straw_map = {
            'trees': [{
                'type': 'root',
                'name': 'cluster',
                'algorithm': 'straw'
            }]
        }

        # Backward compatibility off: fails
        with pytest.raises(RuntimeError) as e:
            LibCrush().parse(straw_map)
        assert 'straw requires backward_compatibility' in str(e.value)

        # Backward compatibility on: it works
        LibCrush(backward_compatibility=True).parse(straw_map)

    def test_parse_trees_invalid_key(self):
        wrong = {
            'trees': [{
                'type': 'root',
                'name': 'dc1',
                1: 'some',
            }]
        }
        with pytest.raises(TypeError):
            LibCrush().parse(wrong)

    def test_parse_invalid_children(self):
        wrong = {
            'trees': [{
                'type': 'root',
                'name': 'dc1',
                'children': 1,
            }]
        }
        with pytest.raises(RuntimeError) as e:
            LibCrush().parse(wrong)
        assert 'must be a list' in str(e.value)

    def test_parse_invalid_children_element(self):
        wrong = {
            'trees': [{
                'type': 'root',
                'name': 'dc1',
                'children': [1],
            }]
        }
        with pytest.raises(RuntimeError) as e:
            LibCrush().parse(wrong)
        assert 'must be a dict' in str(e.value)

    def test_parse_bucket_invalid_key(self):
        wrong = {
            'trees': [{
                'type': 'root',
                'name': 'dc1',
                'INVALID': 1,
            }]
        }
        with pytest.raises(RuntimeError) as e:
            LibCrush().parse(wrong)
        assert 'INVALID is not' in str(e.value)

    def test_parse_device_invalid_key(self):
        wrong = {
            'trees': [{
                'type': 'root',
                'name': 'dc1',
                'children': [{
                    'id': 1,
                    'name': 'device0',
                    'INVALID': 1,
                }]
            }]
        }
        with pytest.raises(RuntimeError) as e:
            LibCrush().parse(wrong)
        assert "'INVALID' is not" in str(e.value)

    def test_parse_invalid_name(self):
        wrong = {
            'trees': [{
                'type': 'root',
                'name': 1,
            }]
        }
        with pytest.raises(TypeError):
            LibCrush().parse(wrong)

    def test_parse_invalid_type(self):
        wrong = {
            'trees': [{
                'type': 1,
            }]
        }
        with pytest.raises(TypeError):
            LibCrush().parse(wrong)

    def test_parse_invalid_id(self):
        wrong = {
            'trees': [{
                'type': 'root',
                'name': 'dc1',
                'id': "some",
            }]
        }
        with pytest.raises(TypeError):
            LibCrush().parse(wrong)

    def test_parse_device_id_invalid(self):
        wrong = {
            'trees': [{
                'type': 'root',
                'name': 'dc1',
                'children': [{
                    'name': 'device0',
                    'id': -1,
                }],
            }]
        }
        with pytest.raises(RuntimeError) as e:
            LibCrush().parse(wrong)
        assert "must be a positive integer" in str(e.value)

    def test_parse_bucket_id_invalid(self):
        wrong = {
            'trees': [{
                'type': 'root',
                'name': 'dc1',
                'id': 2,
            }]
        }
        with pytest.raises(RuntimeError) as e:
            LibCrush().parse(wrong)
        assert "must be a negative integer" in str(e.value)

    def test_parse_bucket_duplicate_id(self):
        duplicate_id = {
            'trees': [{
                'type': 'root',
                'name': 'dc1',
                'id': -1,
                'children': [
                    {
                        'type': 'host',
                        'name': 'host0',
                        'id': -2,
                    },
                    {
                        'type': 'host',
                        'name': 'host1',
                        'id': -2,
                    }
                ],
            }]
        }
        with pytest.raises(RuntimeError) as e:
            LibCrush().parse(duplicate_id)
        assert ' -17 ' in str(e.value)

    def test_parse_bucket_reference_id(self):
        crushmap = {
            'trees': [{
                'type': 'root',
                'name': 'dc1',
                'id': -1,
                'children': [
                    {
                        'weight': 1 * 0x10000,
                        'reference_id': -2,
                    },
                    {
                        'type': 'host',
                        'name': 'host1',
                        'id': -2,
                    }
                ],
            }]
        }
        LibCrush(verbose=1).parse(crushmap)

    def test_parse_bucket_reference_id_bad(self):
        crushmap = {
            'trees': [{
                'type': 'root',
                'name': 'dc1',
                'id': -1,
                'children': [
                    {
                        'weight': 1 * 0x10000,
                        'reference_id': 'bad',
                    }
                ],
            }]
        }
        with pytest.raises(TypeError):
            LibCrush(verbose=1).parse(crushmap)

        crushmap = {
            'trees': [{
                'type': 'root',
                'name': 'dc1',
                'id': -1,
                'children': [
                    {
                        'weight': 1 * 0x10000,
                        'reference_id': 1,
                        'INVALID': 'foo',
                    }
                ],
            }]
        }
        with pytest.raises(RuntimeError) as e:
            LibCrush().parse(crushmap)
        assert 'INVALID is not among' in str(e.value)

    def test_parse_weight_invalid(self):
        wrong = {
            'trees': [{
                'type': 'root',
                'name': 'dc1',
                'weight': "some",
            }]
        }
        with pytest.raises(RuntimeError) as e:
            LibCrush().parse(wrong)
        assert 'must be an int' in str(e.value)

    def test_parse_rules_type_wrong(self):
        wrong = {
            'rules': []
        }
        with pytest.raises(RuntimeError) as e:
            LibCrush().parse(wrong)
        assert 'must be a dict' in str(e.value)

    def test_parse_rules_invalid_key(self):
        wrong = {
            'rules': {
                1: 'root',
            }
        }
        with pytest.raises(TypeError):
            LibCrush().parse(wrong)

    def test_parse_step_missing_operand(self):
        wrong = {
            'rules': {
                'data': [
                    []
                ]
            }
        }
        with pytest.raises(RuntimeError) as e:
            LibCrush().parse(wrong)
        assert 'missing operand' in str(e.value)

    def test_parse_step_unknown_operand(self):
        wrong = {
            'rules': {
                'data': [
                    ['unknown']
                ]
            }
        }
        with pytest.raises(RuntimeError) as e:
            LibCrush().parse(wrong)
        assert 'operand unknown' in str(e.value)

    def test_parse_step_operand_bad_type(self):
        wrong = {
            'rules': {
                'data': [
                    [1]
                ]
            }
        }
        with pytest.raises(TypeError):
            LibCrush().parse(wrong)

    def test_parse_step_take_bad(self):
        wrong = {
            'rules': {
                'data': [
                    ["take"]
                ]
            }
        }
        with pytest.raises(RuntimeError) as e:
            LibCrush().parse(wrong)
        assert 'exactly two' in str(e.value)

        wrong = {
            'rules': {
                'data': [
                    ["take", 1]
                ]
            }
        }
        with pytest.raises(TypeError) as e:
            LibCrush().parse(wrong)

        wrong = {
            'rules': {
                'data': [
                    ["take", u"unknown"]
                ]
            }
        }
        with pytest.raises(RuntimeError) as e:
            LibCrush().parse(wrong)
        assert 'not a known bucket' in str(e.value)

    def test_parse_step_emit_bad(self):
        wrong = {
            'rules': {
                'data': [
                    ["emit", 2]
                ]
            }
        }
        with pytest.raises(RuntimeError) as e:
            LibCrush().parse(wrong)
        assert 'exactly one' in str(e.value)

    def test_parse_step_set_backward(self):
        for backward in STEP_BACKWARDS:
            crushmap = {
                'rules': {
                    'data': [
                        ["set_" + backward, 1234]
                    ]
                }
            }
            with pytest.raises(RuntimeError) as e:
                LibCrush(verbose=1).parse(crushmap)
            assert 'not allowed unless backward_compatibility' in str(e.value)
            LibCrush(verbose=1, backward_compatibility=1).parse(crushmap)

    def test_parse_step_set_bad(self):
        wrong = {
            'rules': {
                'data': [
                    ["set_choose_tries"]
                ]
            }
        }
        with pytest.raises(RuntimeError) as e:
            LibCrush().parse(wrong)
        assert 'exactly two' in str(e.value)

        wrong = {
            'rules': {
                'data': [
                    ["set_wrong", 1]
                ]
            }
        }
        with pytest.raises(RuntimeError) as e:
            LibCrush().parse(wrong)
        assert 'set operand unknown' in str(e.value)

        wrong = {
            'rules': {
                'data': [
                    ["set_choose_tries", "LK"]
                ]
            }
        }
        with pytest.raises(Exception) as e:
            LibCrush().parse(wrong)

    def test_parse_step_choose(self):
        wrong = {
            'rules': {
                'data': [
                    ["choose"]
                ]
            }
        }
        with pytest.raises(RuntimeError) as e:
            LibCrush().parse(wrong)
        assert 'exactly five' in str(e.value)

        wrong = {
            'rules': {
                'data': [
                    ["choose", 1, 2, 3, 4]
                ]
            }
        }
        with pytest.raises(TypeError) as e:
            LibCrush().parse(wrong)

        wrong = {
            'rules': {
                'data': [
                    ["choose", "notgood", 2, 3, 4]
                ]
            }
        }
        with pytest.raises(RuntimeError) as e:
            LibCrush().parse(wrong)
        assert 'unknown notgood' in str(e.value)

        wrong = {
            'rules': {
                'data': [
                    ["choose", "firstn", "notgood", 3, 4]
                ]
            }
        }
        with pytest.raises(TypeError) as e:
            LibCrush().parse(wrong)

        wrong = {
            'rules': {
                'data': [
                    ["choose", "firstn", 0, "notgood", 4]
                ]
            }
        }
        with pytest.raises(RuntimeError) as e:
            LibCrush().parse(wrong)
        assert "must be 'type'" in str(e.value)

        wrong = {
            'rules': {
                'data': [
                    ["choose", "firstn", 0, "type", "unknown"]
                ]
            }
        }
        with pytest.raises(RuntimeError) as e:
            LibCrush().parse(wrong)
        assert "type is unknown" in str(e.value)

    def test_parse_choose_args(self):
        wrong = {
            'choose_args': 0
        }
        with pytest.raises(RuntimeError) as e:
            LibCrush().parse(wrong)
        assert "must be a dict" in str(e.value)

        wrong = {
            'choose_args': {
                "1": 0
            }
        }
        with pytest.raises(RuntimeError) as e:
            LibCrush().parse(wrong)
        assert "must be a list" in str(e.value)

        wrong = {
            'choose_args': {
                "1": [
                    0,
                ]
            }
        }
        with pytest.raises(RuntimeError) as e:
            LibCrush().parse(wrong)
        assert "must be a dict" in str(e.value)

        wrong = {
            'choose_args': {
                "1": [
                    {
                    },
                ]
            }
        }
        with pytest.raises(RuntimeError) as e:
            LibCrush().parse(wrong)
        assert "bucket_id or bucket_name are required" in str(e.value)

        wrong = {
            'choose_args': {
                "1": [
                    {
                        "bucket_id": []
                    },
                ]
            }
        }
        with pytest.raises(TypeError) as e:
            LibCrush().parse(wrong)

        wrong = {
            'choose_args': {
                "1": [
                    {
                        "bucket_id": 5
                    },
                ]
            }
        }
        with pytest.raises(RuntimeError) as e:
            LibCrush().parse(wrong)
        assert "id must be a negative integer" in str(e.value)

        wrong = {
            'choose_args': {
                "1": [
                    {
                        "bucket_name": "unknown"
                    },
                ]
            }
        }
        with pytest.raises(RuntimeError) as e:
            LibCrush().parse(wrong)
        assert "not a known bucket" in str(e.value)

        wrong = {
            'choose_args': {
                "1": [
                    {
                        "bucket_id": -3
                    },
                ]
            }
        }
        with pytest.raises(RuntimeError) as e:
            LibCrush().parse(wrong)
        assert "not in [0,0[" in str(e.value)

        wrong = {
            'trees': [{
                'type': 'root',
                'name': 'dc1',
                'id': -1,
                'children': [],
            }],
            'choose_args': {
                "1": [
                    {
                        "bucket_id": -1,
                        "ids": 0
                    },
                ]
            }
        }
        with pytest.raises(RuntimeError) as e:
            LibCrush().parse(wrong)
        assert "must be a list" in str(e.value)

        wrong = {
            'trees': [{
                'type': 'root',
                'name': 'dc1',
                'id': -1,
                'children': [],
            }],
            'choose_args': {
                "1": [
                    {
                        "bucket_id": -1,
                        "weight_set": 0
                    },
                ]
            }
        }
        with pytest.raises(RuntimeError) as e:
            LibCrush().parse(wrong)
        assert "must be a list" in str(e.value)

        wrong = {
            'trees': [{
                'type': 'root',
                'name': 'dc1',
                'id': -1,
                'children': [],
            }],
            'choose_args': {
                "1": [
                    {
                        "bucket_id": -1,
                        "ids": [50]
                    },
                ]
            }
        }
        with pytest.raises(RuntimeError) as e:
            LibCrush().parse(wrong)
        assert "expected a list of ids with 0 elements and got 1 instead" in str(e.value)

        wrong = {
            'trees': [{
                'type': 'root',
                'name': 'dc1',
                'id': -1,
                'children': [],
            }],
            'choose_args': {
                "1": [
                    {
                        "bucket_id": -1,
                        "weight_set": [[10 * 0x10000]]
                    },
                ]
            }
        }
        with pytest.raises(RuntimeError) as e:
            LibCrush().parse(wrong)
        assert "expected a list of weights with 0 elements and got 1 instead" in str(e.value)

        wrong = {
            'trees': [{
                'type': 'root',
                'name': 'dc1',
                'id': -1,
                'children': [],
            }],
            'choose_args': {
                "1": [
                    {
                        "bucket_id": -1,
                        "weight_set": [0]
                    },
                ]
            }
        }
        with pytest.raises(RuntimeError) as e:
            LibCrush().parse(wrong)
        assert "must be a list" in str(e.value)

        wrong = {
            'trees': [{
                'type': 'root',
                'name': 'dc1',
                'id': -1,
                'children': [
                    {
                        "id": 0,
                        "name": "device1",
                    }
                ],
            }],
            'choose_args': {
                "1": [
                    {
                        "bucket_id": -1,
                        "weight_set": [["bad"]]
                    },
                ]
            }
        }
        with pytest.raises(RuntimeError) as e:
            LibCrush().parse(wrong)
        assert "must be an int" in str(e.value)

    def test_map_ok_choose_args(self):
        crushmap = {
            "trees": [
                {
                    "type": "root",
                    "id": -1,
                    "name": "dc1",
                    "children": [],
                }
            ],
            "rules": {
                "data": [
                    ["take", "dc1"],
                    ["chooseleaf", "firstn", 0, "type", "host"],
                    ["emit"]
                ],
                "for_validation": [
                    ["take", "dc1"],
                    ["chooseleaf", "firstn", 0, "type", 0],
                    ["emit"]
                ],
            }
        }
        crushmap['trees'][0]['children'].extend([
            {
                "type": "host",
                "id": -(i + 2),
                "name": "host%d" % i,
                "children": [
                    {"id": (2 * i), "name": "device%02d" % (2 * i), "weight": 1 * 0x10000},
                    {"id": (2 * i + 1), "name": "device%02d" % (2 * i + 1), "weight": 2 * 0x10000},
                ],
            } for i in range(0, 10)
        ])
        crushmap['choose_args'] = {
            "1": [
                {
                    "bucket_name": "host9",
                    "weight_set": [[2 * 0x10000, 1 * 0x10000]]  # invert the weights
                },
            ]
        }
        c = LibCrush(verbose=1)
        assert c.parse(crushmap)
        assert c.map(rule="data",
                     value=1234,
                     replication_count=2) == ["device19", "device13"]
        assert c.map(rule="data",
                     value=1234,
                     replication_count=2,
                     choose_args="1") == ["device18", "device13"]
        assert c.map(rule="data",
                     value=1234,
                     replication_count=2,
                     choose_args=crushmap['choose_args']["1"]) == ["device18", "device13"]

    def test_map_ok(self):
        crushmap = {
            "trees": [
                {
                    "type": "root",
                    "id": -1,
                    "name": "dc1",
                    "children": [],
                }
            ],
            "rules": {
                "data": [
                    ["take", "dc1"],
                    ["chooseleaf", "firstn", 0, "type", "host"],
                    ["emit"]
                ],
                "for_validation": [
                    ["take", "dc1"],
                    ["chooseleaf", "firstn", 0, "type", 0],
                    ["emit"]
                ],
            }
        }
        crushmap['trees'][0]['children'].extend([
            {
                "type": "host",
                "id": -(i + 2),
                "name": "host%d" % i,
                "children": [
                    {"id": (2 * i), "name": "device%02d" % (2 * i), "weight": 1 * 0x10000},
                    {"id": (2 * i + 1), "name": "device%02d" % (2 * i + 1), "weight": 2 * 0x10000},
                ],
            } for i in range(0, 10)
        ])
        c = LibCrush(verbose=1)
        assert c.parse(crushmap)
        assert c.map(rule="data",
                     value=1234,
                     replication_count=1) == ["device19"]
        assert c.map(rule="data",
                     value=1234,
                     replication_count=2) == ["device19", "device13"]

    def test_map_bad(self):
        crushmap = {
            "trees": [
                {
                    "type": "root",
                    "id": -1,
                    "name": "dc1",
                    "children": [
                        {
                            "id": -2,
                            "name": "host0",
                            "type": "host",
                            "children": [
                                {"id": 0, "name": "device0", "weight": 1 * 0x10000},
                                {"id": 1, "name": "device1", "weight": 2 * 0x10000},
                            ]
                        }
                    ],
                }
            ],
            "rules": {
                "firstn": [
                    ["take", "dc1"],
                    ["chooseleaf", "firstn", 0, "type", "host"],
                    ["emit"]
                ],
                "indep": [
                    ["take", "dc1"],
                    ["chooseleaf", "indep", 0, "type", "host"],
                    ["emit"]
                ]
            }
        }
        c = LibCrush(verbose=1)
        assert c.parse(crushmap)
        assert c.map(rule="firstn",
                     value=1234,
                     replication_count=2) == ["device1"]
        assert c.map(rule="indep",
                     value=1234,
                     replication_count=2) == ["device1", None]

    def test_map_missing_map(self):
        with pytest.raises(RuntimeError) as e:
            LibCrush().map(rule="data", value=1234, replication_count=2)
        assert 'call parse()' in str(e.value)

    def test_map_no_rule(self):
        c = LibCrush(verbose=1)
        assert c.parse({})
        with pytest.raises(RuntimeError) as e:
            c.map(rule="norule",
                  value=1234,
                  replication_count=1)
        assert 'norule is not found' in str(e.value)

    def test_map_bad_replication_count(self):
        c = LibCrush(verbose=1)
        assert c.parse({
            "rules": {"data": []}
        })
        with pytest.raises(RuntimeError) as e:
            c.map(rule="data",
                  value=1234,
                  replication_count=0)
        assert 'must be >= 1' in str(e.value)

    def test_map_wrong_weights(self):
        c = LibCrush(verbose=1)
        assert c.parse({
            "rules": {"data": []}
        })
        with pytest.raises(RuntimeError) as e:
            c.map(rule="data",
                  value=1234,
                  replication_count=1,
                  weights={"unknowndevice": 1 * 0x10000})
        assert 'unknowndevice is not a known device' in str(e.value)

        assert c.parse({
            "trees": [{
                "type": "root",
                "name": "dc1",
                "children": [{
                    "name": "device0",
                    "id": 0
                }]
            }],
            "rules": {"data": []}
        })
        with pytest.raises(TypeError) as e:
            c.map(rule="data",
                  value=1234,
                  replication_count=1,
                  weights={"device0": "abc"})

    def test_map_fail(self):
        crushmap = {
            "trees": [{
                "type": "root",
                "name": "dc1",
                "id": -1,
                "children": [{
                    "type": "host",
                    "name": "host0",
                    "id": -2
                }]
            }],
            "rules": {
                "data": [
                    ["take", "dc1"],
                    ["chooseleaf", "firstn", 0, "type", "host"],
                    ["emit"]
                ]
            }
        }
        c = LibCrush(verbose=1)
        assert c.parse(crushmap)
        with pytest.raises(RuntimeError) as e:
            c.map(rule="data",
                  value=1234,
                  replication_count=1)
        assert 'unable to map' in str(e.value)

    def test_convert(self):
        c = LibCrush(verbose=1)
        crushmap = c.ceph_read("tests/sample-ceph-crushmap.txt")
        assert 'devices' in crushmap
        crushmap = c.ceph_read("tests/sample-ceph-crushmap.crush")
        assert 'devices' in crushmap

    def test_pool_pps(self):
        c = LibCrush()

        pps_1 = c.ceph_pool_pps(0, 16, 16)
        assert 430787817 == pps_1['0.0']
        pps_1_values = sorted(set(pps_1.values()))
        assert 16 == len(pps_1)

        pps_2 = c.ceph_pool_pps(0, 23, 16)
        pps_2_values = sorted(set(pps_2.values()))
        assert pps_1_values == pps_2_values

    def test_ceph_incompat(self):
        c = LibCrush(verbose=1)

        assert c.ceph_incompat() is False

        c.parse({})
        assert c.ceph_incompat() is False

        c.parse({'choose_args': {
            1: [],
            2: [],
        }})
        assert c.ceph_incompat() is True

        c.parse({'choose_args': {
            1: [],
        }})
        assert c.ceph_incompat() is False

        with pytest.raises(RuntimeError) as e:
            c.parse({'choose_args': {
                "1": [],
            }})
            assert c.ceph_incompat() is False
        assert 'Invalid argument' in str(e.value)

# Local Variables:
# compile-command: "cd .. ; tox -e py27 -- -s -vv tests/test_libcrush.py"
# End:



# -*- mode: python; coding: utf-8 -*-
#
# Copyright (C) 2017 <contact@redhat.com>
#
# Author: Loic Dachary <loic@dachary.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
import argparse
import pickle
import pprint

from crush import analyze
from crush import main
from crush import optimize


class TestCrush(object):

    def test_run(self):
        # needs a bit more love to actually work
        pass
        # c = main.Crush()
        # argv = []
        # c.run(argv)
        # assert (logging.getLogger('crush').getEffectiveLevel() ==
        #         logging.INFO)
        # c.run(['--verbose'] + argv)
        # assert (logging.getLogger('crush').getEffectiveLevel() ==
        #         logging.DEBUG)

    def test_pickle(self):
        m = main.Main()
        m.parse(['optimize'])
        p = pickle.dumps(m)
        n = pickle.loads(p)
        assert n.argv == m.argv

    def test_get_trimmed_argv(self):
        d_parser = optimize.Optimize.get_parser()
        parser = argparse.ArgumentParser(
            parents=[
                d_parser,
            ],
            conflict_handler='resolve',
        )
        parser.add_argument('--something-with-arg')
        parser.add_argument('--no-arg', action='store_true')
        pprint.pprint(vars(parser))
        argv_discarded = [
            '--no-arg',
            '--something-with-arg', 'arg',
            '--step', '200',
        ]
        argv_preserved = [
            '--type', 'device',
        ]
        argv = argv_discarded + argv_preserved
        args = parser.parse_args(argv)
        trimmed_argv = main.Main.get_trimmed_argv(
            analyze.Analyze.get_parser(), args)
        expected = ['--type', 'device']
        assert expected == trimmed_argv


# Local Variables:
# compile-command: "cd .. ; tox -e py27 -- -s -vv tests/test_main.py"
# End:



# -*- mode: python; coding: utf-8 -*-
#
# Copyright (C) 2017 <contact@redhat.com>
#
# Author: Loic Dachary <loic@dachary.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from crush.ceph import Ceph


class TestAnalyze(object):

    def test_report_compat(self):
        #
        # verify --choose-args is set to the pool when the crushmap contains
        # *-target-weights buckets.
        #
        for p in ([],
                  ['--choose-args=3'],
                  ['--pool=3'],
                  ['--choose-args=3', '--pool=3']):
            a = Ceph().constructor([
                '--verbose',
                'analyze',
                '--crushmap', 'tests/ceph/ceph-report-compat.json',
            ] + p)
            d = a.analyze_report(*a.analyze())
            print(str(d))
            expected = """\
        ~id~  ~weight~  ~PGs~  ~over/under filled %~
~name~                                              
host0     -1     65536      1                    0.0
host1     -2     65536      1                    0.0
host2     -5     65536      1                    0.0

Worst case scenario if a host fails:

        ~over filled %~
~type~                 
device            33.33
host              33.33
rack               0.00
root               0.00\
""" # noqa trailing whitespaces are expected
            assert expected == str(d)

# Local Variables:
# compile-command: "cd .. ; tox -e py27 -- -s -vv tests/test_ceph_analyze.py"
# End:



# -*- mode: python; coding: utf-8 -*-
#
# Copyright (C) 2017 <contact@redhat.com>
#
# Author: Loic Dachary <loic@dachary.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import division

import argparse
import collections
import pandas as pd
import logging
import textwrap

from crush import Crush

log = logging.getLogger(__name__)


class Compare(object):

    orig_weights = None
    dest_weights = None

    def __init__(self, args, main):
        self.args = args
        if self.args.choose_args and self.args.destination_choose_args is None:
            self.args.destination_choose_args = self.args.choose_args
        if self.args.choose_args and self.args.origin_choose_args is None:
            self.args.origin_choose_args = self.args.choose_args
        self.main = main

    def set_origin(self, c):
        self.origin = c

    def set_origin_crushmap(self, origin):
        o = Crush(backward_compatibility=self.args.backward_compatibility)
        o.parse(self.main.convert_to_crushmap(origin))
        self.set_origin(o)

    def set_destination(self, c):
        self.destination = c

    def set_destination_crushmap(self, destination):
        d = Crush(backward_compatibility=self.args.backward_compatibility)
        d.parse(self.main.convert_to_crushmap(destination))
        self.set_destination(d)

    @staticmethod
    def get_parser():
        parser = argparse.ArgumentParser(
            add_help=False,
            conflict_handler='resolve',
        )
        replication_count = 3
        parser.add_argument(
            '--replication-count',
            help=('number of devices to map (default: %d)' % replication_count),
            type=int,
            default=replication_count)
        parser.add_argument(
            '--rule',
            help='the name of rule')
        parser.add_argument(
            '--choose-args',
            help='modify the origin and destination weights')
        parser.add_argument(
            '--origin-choose-args',
            help='modify the origin weights')
        parser.add_argument(
            '--destination-choose-args',
            help='modify the destination weights')
        values_count = 100000
        parser.add_argument(
            '--values-count',
            help='repeat mapping (default: %d)' % values_count,
            type=int,
            default=values_count)
        parser.add_argument(
            '--origin',
            metavar='PATH',
            help='PATH to the origin crushmap file')
        parser.add_argument(
            '--destination',
            metavar='PATH',
            help='PATH to the destination crushmap file')
        parser.add_argument(
            "-ow", "--origin-weights",
            help="Weights file to apply to the origin map")
        parser.add_argument(
            "-dw", "--destination-weights",
            help="Weights file to apply to the destination map")
        parser.add_argument(
            '--order-matters',
            action='store_true', default=False,
            help='true if the order of mapped devices matter (default: false)')
        return parser

    @staticmethod
    def set_parser(subparsers, arguments):
        parser = Compare.get_parser()
        arguments(parser)
        subparsers.add_parser(
            'compare',
            formatter_class=argparse.RawDescriptionHelpFormatter,
            description=textwrap.dedent("""\
            Compare crushmaps

            After a crushmap is changed (e.g. addition/removal of
            items, modification of weights or tunables), objects may
            move from an item to another.

            The crushmap before the modification is specified with the
            --origin option and the crushmap after the modification is
            specified with the --destination option.

            The format of the crushmap file specified with --origin or
            --destination can either be:

            - a JSON representation of a crushmap as documented in the
              Crush.parse_crushmap() method

            - a Ceph binary, text or JSON crushmap compatible with
              Luminuous and below

            Each crushmap is given the same set of objects (in the
            range [0,--value-count[) to map using a given rule
            (--rule) with a given replication count
            (--replication-count). If an object is mapped to the same
            items by both crushmaps, it does not move. If it is mapped
            to different items, it moves.

            The order in which items are mapped may or may not matter.
            For instance, it does not matter if three exact copies of
            an object are mapped to items 23, 78 and 45 instead of to
            45, 78 and 23. But it matters in the case of erasure-coded
            objects because they are split among multiple items. A
            simple example: first half of object is stored in one
            item, second half is stored in a second item, and parity
            is stored in a third. A rule with a "choose*" "indep" step
            is typically used when the order of the mapping order matters
            and a rule with a "choose*" "firstn" step is most often
            used when the mapping does not matter.

            If an object maps to items 23, 78 and 45 in the --origin
            crushmap and to 45, 78 and 23 in the --destination crushmap
            (i.e. the same items but in a different order), it is assumed
            the order does not matter and the object did not move. If
            the --order-matters flag is set, it is assumed that some object
            movement will be necessary and the object will be counted as
            moving from 45 to 23 and from 23 to 45.

            """),
            epilog=textwrap.dedent("""
            Examples:

            $ crush compare --rule firstn \\
                            --replication-count 1 \\
                            --origin before.json --destination after.json
            There are 1000 objects.

            Replacing the crushmap specified with --origin with the crushmap
            specified with --destination will move 229 objects (22.9% of the total)
            from one item to another.

            The rows below show the number of objects moved from the
            given item to each item named in the columns. The objects%
            at the end of the rows shows the percentage of the total
            number of objects that is moved away from this particular
            item. The last row shows the percentage of the total
            number of objects that is moved to the item named in the
            column.

                     osd.8    osd.9    objects%
            osd.0        3        4       0.70%
            osd.1        1        3       0.40%
            osd.2       16       16       3.20%
            osd.3       19       21       4.00%
            osd.4       17       18       3.50%
            osd.5       18       23       4.10%
            osd.6       14       23       3.70%
            osd.7       14       19       3.30%
            objects%   10.20%   12.70%   22.90%

            """),
            help='Compare crushmaps',
            parents=[parser],
        ).set_defaults(
            func=Compare,
        )

    def compare(self):
        a = self.origin
        self.origin_d = collections.defaultdict(lambda: 0)
        b = self.destination
        self.destination_d = collections.defaultdict(lambda: 0)
        replication_count = self.args.replication_count
        values = self.main.hook_create_values()
        rule = self.args.rule
        self.from_to = collections.defaultdict(lambda: collections.defaultdict(lambda: 0))
        for (name, value) in values.items():
            am = a.map(rule, value, replication_count, self.orig_weights,
                       choose_args=self.args.origin_choose_args)
            log.debug("am {} == {} mapped to {}".format(name, value, am))
            assert len(am) == replication_count
            for d in am:
                self.origin_d[d] += 1
            bm = b.map(rule, value, replication_count, self.dest_weights,
                       choose_args=self.args.destination_choose_args)
            log.debug("bm {} == {} mapped to {}".format(name, value, bm))
            assert len(bm) == replication_count
            for d in bm:
                self.destination_d[d] += 1
            if self.args.order_matters:
                for i in range(len(am)):
                    if am[i] != bm[i]:
                        self.from_to[am[i]][bm[i]] += 1
            else:
                am = set(am)
                bm = set(bm)
                if am == bm:
                    continue
                ar = list(am - bm)
                br = list(bm - am)
                for i in range(len(ar)):
                    self.from_to[ar[i]][br[i]] += 1
        return self.from_to

    def compare_bucket(self, bucket):
        a = self.origin
        self.origin_d = collections.defaultdict(lambda: 0)
        b = self.destination
        self.destination_d = collections.defaultdict(lambda: 0)
        replication_count = self.args.replication_count
        values = self.main.hook_create_values()
        rule = self.args.rule
        self.from_to = collections.defaultdict(lambda: collections.defaultdict(lambda: 0))
        self.in_out = collections.defaultdict(lambda: collections.defaultdict(lambda: 0))
        item2path = a.collect_item2path([bucket])
        log.debug("item2path " + str(item2path))
        for (name, value) in values.items():
            am = a.map(rule, value, replication_count, self.orig_weights,
                       choose_args=self.args.choose_args)
            log.debug("am {} == {} mapped to {}".format(name, value, am))
            assert len(am) == replication_count
            bm = b.map(rule, value, replication_count, self.dest_weights,
                       choose_args=self.args.choose_args)
            log.debug("bm {} == {} mapped to {}".format(name, value, bm))
            assert len(bm) == replication_count
            if self.args.order_matters:
                for i in range(len(am)):
                    if am[i] != bm[i]:
                        a_path = item2path.get(am[i])
                        b_path = item2path.get(bm[i])
                        if a_path is None and b_path is None:
                            continue
                        if a_path is None or b_path is None:
                            self.in_out[am[i]][bm[i]] += 1
                            continue
                        self.from_to[a_path[1]][b_path[1]] += 1
            else:
                am = set(am)
                bm = set(bm)
                if am == bm:
                    continue
                ar = sorted(list(am - bm))
                br = sorted(list(bm - am))
                for i in range(len(ar)):
                    a_path = item2path.get(ar[i])
                    b_path = item2path.get(br[i])
                    if a_path is None and b_path is None:
                        continue
                    if a_path is None or b_path is None:
                        self.in_out[ar[i]][br[i]] += 1
                        continue
                    self.from_to[a_path[1]][b_path[1]] += 1
        return (self.from_to, self.in_out)

    def display(self):
        out = ""
        o = pd.Series(self.origin_d)
        objects_count = o.sum()
        out += "There are {} objects.\n".format(objects_count)
        m = pd.DataFrame.from_dict(self.from_to, dtype=int).fillna(0).T.astype(int)
        objects_moved = m.sum().sum()
        objects_moved_percent = objects_moved / objects_count * 100
        out += textwrap.dedent("""
        Replacing the crushmap specified with --origin with the crushmap
        specified with --destination will move {} objects ({}% of the total)
        from one item to another.
        """.format(int(objects_moved), objects_moved_percent))
        from_to_percent = m.sum(axis=1) / objects_count
        to_from_percent = m.sum() / objects_count
        n = self.main.value_name()
        m[n + '%'] = from_to_percent.apply(lambda v: "{:.2%}".format(v))
        mt = m.T
        mt[n + '%'] = to_from_percent.apply(lambda v: "{:.2%}".format(v))
        m = mt.T.fillna("{:.2f}%".format(objects_moved_percent))
        out += textwrap.dedent("""
        The rows below show the number of objects moved from the given
        item to each item named in the columns. The objects% at the
        end of the rows shows the percentage of the total number
        of objects that is moved away from this particular item. The
        last row shows the percentage of the total number of objects
        that is moved to the item named in the column.

        """)
        pd.set_option('display.max_rows', None)
        pd.set_option('display.width', 160)
        out += str(m)
        return out

    def run(self):
        self.run_compare()
        print(self.display())

    def run_compare(self):
        self.set_origin_crushmap(self.args.origin)
        self.set_destination_crushmap(self.args.destination)

        if self.args.origin_weights:
            with open(self.args.origin_weights) as f_ow:
                self.orig_weights = Crush.parse_weights_file(f_ow)
        if self.args.destination_weights:
            with open(self.args.destination_weights) as f_dw:
                self.dest_weights = Crush.parse_weights_file(f_dw)

        self.compare()
